# -*- coding: utf-8 -*-
"""The core module for the api_24sea.datasignals subpackage"""
import asyncio
import datetime
import itertools
import logging
from typing import Any, Dict, List, Optional, Union
from warnings import simplefilter

import httpx
import pandas as pd
from pandas import __version__ as pd_version

from .. import exceptions as E
from .. import utils as U

# Local imports
from ..abc import AuthABC
from . import schemas as S

# This filter is used to ignore the PerformanceWarning that is raised when
# the DataFrame is modified in place. This is the case when we add columns
# to the DataFrame in the get_data method.
# This is the only way to update the DataFrame in place when using accessors
# and performance is not an issue in this case.
# See https://stackoverflow.com/a/76306267/7169710 for reference.
simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

logging.basicConfig(format="%(message)s", level=logging.WARNING)


class API(AuthABC):
    """Accessor for working with data signals coming from the 24SEA API."""

    def __init__(self):
        super().__init__()
        self._metrics_overview = self._permissions_overview
        self.base_url: str = str(f"{self.base_url}datasignals/")
        self._selected_metrics: Optional[pd.DataFrame] = None

    @property
    def authenticated(self) -> bool:
        """Whether the client is authenticated"""
        return self._authenticated

    @property
    @U.require_auth
    def metrics_overview(self) -> Optional[pd.DataFrame]:
        """Get the metrics overview DataFrame."""
        return self._metrics_overview

    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def authenticate(
        self,
        username: str,
        password: str,
        permissions_overview: Optional[pd.DataFrame] = None,
    ):
        """Authenticate with username/password"""
        self._username = username
        self._password = password
        self._metrics_overview = (
            permissions_overview
            if permissions_overview is not None
            else self._metrics_overview
        )
        self._auth = httpx.BasicAuth(self._username, self._password)

        r_profile = U.handle_request(
            f"{self.base_url}profile/",
            {"username": self._username},
            self._auth,
            {"accept": "application/json"},
        )
        if (
            r_profile.status_code == 200 or self._metrics_overview is not None
        ):  # noqa: E501
            self._authenticated = True
        # fmt: off
        logging.info(f"\033[32;1m{username} has access to "
                        f"\033[4m{U.BASE_URL}.\033[0m")


        if self._metrics_overview is not None:
            return self

        logging.info("Now getting your metrics_overview table...")
        r_metrics = U.handle_request(
            f"{self.base_url}metrics/",
            {"project": None, "locations": None, "metrics": None},
            self._auth,
            {"accept": "application/json"},
        )
        # fmt: off
        if not isinstance(r_metrics, type(None)):
            try:
                m_ = pd.DataFrame(r_metrics.json())
            except Exception:
                raise E.ProfileError(f"\033[31;1mThe metrics overview is empty. This is your profile information:"  # noqa: E501  # pylint: disable=C0301
                                     f"\n {r_profile.json()}")
            if m_.empty:
                raise E.ProfileError(f"\033[31;1mThe metrics overview is empty. This is your profile information:"  # noqa: E501  # pylint: disable=C0301
                                    f"\n {r_profile.json()}")
            try:
                s_ = m_.apply(lambda x: x["metric"]
                            .replace(x["statistic"], "")
                            .replace(x["short_hand"], "")
                            .strip(), axis=1).str.strip("_").str.split("_", expand=True)  # noqa: E501  # pylint: disable=C0301
                # Just take the first two columns to avoid duplicates
                s_ = s_.iloc[:, :2]
                s_.columns = ["site_id", "location_id"]
            except Exception:
                self._metrics_overview = m_
                return self
        # fmt: on

            self._metrics_overview = pd.concat([m_, s_], axis=1)
        return self

    @U.require_auth
    @U.validate_call
    def get_metrics(
        self,
        site: Optional[str] = None,
        locations: Optional[Union[str, List[str]]] = None,
        metrics: Optional[Union[str, List[str]]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Optional[List[Dict[str, Optional[str]]]]:
        """
        Get the metrics names for a site, provided the following parameters.

        Parameters
        ----------
        site : Optional[str]
            The site name. If None, the queryable metrics for all sites
            will be returned, and the locations and metrics parameters will be
            ignored.
        locations : Optional[Union[str, List[str]]]
            The locations for which to get the metrics. If None, all locations
            will be considered.
        metrics : Optional[Union[str, List[str]]]
            The metrics to get. They can be specified as regular expressions.
            If None, all metrics will be considered.

            For example:

            * | ``metrics=["^ACC", "^DEM"]`` will return all the metrics that
              | start with ACC or DEM,
            * Similarly, ``metrics=["windspeed$", "winddirection$"]`` will
              | return all the metrics that end with windspeed and
              | winddirection,
            * and ``metrics=[".*WF_A01.*",".*WF_A02.*"]`` will return all
              | metrics that contain WF_A01 or WF_A02.

        Returns
        -------
        Optional[List[Dict[str, Optional[str]]]]
            The metrics names for the given site, locations and metrics.

        .. note::
            This class method is legacy because it does not add functionality to
            the DataSignals pandas accessor.

        """
        url = f"{self.base_url}metrics/"
        if headers is None:
            headers = {"accept": "application/json"}
        if site is None:
            params = {}
        if isinstance(locations, List):
            locations = ",".join(locations)
        if isinstance(metrics, List):
            metrics = ",".join(metrics)
        params = {
            "project": site,
            "locations": locations,
            "metrics": metrics,
        }

        r_ = U.handle_request(url, params, self._auth, headers)

        # Set the return type of the get_metrics method to the Metrics schema
        return r_.json()  # type: ignore

    @U.require_auth
    def selected_metrics(self, data: pd.DataFrame) -> pd.DataFrame:
        """Return the selected metrics for the query."""
        if self._metrics_overview is None:
            raise E.ProfileError(
                "\033[31mThe metrics overview is empty. Please authenticate "
                "first with the \033[1mauthenticate\033[22m method."
            )
        if data.empty:
            raise E.DataSignalsError(
                "\033[31mThe \033[1mselected_metrics\033[22m method can only "
                "be called if the DataFrame is not empty, or after the "
                "\033[1mget_data\033[22m method has been called."
            )
        # Get the selected metrics as the Data columns that are available
        # in the metrics_overview DataFrame
        return self._metrics_overview[
            self._metrics_overview["metric"].isin(data.columns)
        ].set_index("metric")

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_data(
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        as_star_schema: bool = False,
        outer_join_on_timestamp: bool = True,
        headers: Optional[Dict[str, str]] = None,
        data: Optional[pd.DataFrame] = None,
        timeout: int = 3600,
        threads: Optional[int] = None,
    ) -> Optional[
        Union[
            pd.DataFrame,
            Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]],
            List[Union[Any, str]],
        ]
    ]:
        """Get the data signals from the 24SEA API.

        Parameters
        ----------
        sites : Optional[Union[List, str]]
            The site name or List of site names. If None, the site will be
            inferred from the metrics.
        locations : Optional[Union[List, str]]
            The location name or List of location names. If None, the location
            will be inferred from the metrics.
        metrics : Union[List, str]
            The metric name or List of metric names. It must be provided.
            They do not have to be the entire metric name, but can be a part
            of it. For example, if the metric name is
            ``"mean_WF_A01_windspeed"``, the user can equivalently provide
            ``sites="wf"``, ``locations="a01"``, ``metric="mean windspeed"``.
        start_timestamp : Union[str, datetime.datetime]
            The start timestamp for the query. It must be in ISO 8601 format,
            e.g., ``"2021-01-01T00:00:00Z"`` or a datetime object.
        end_timestamp : Union[str, datetime.datetime]
            The end timestamp for the query. It must be in ISO 8601 format,
            e.g., ``"2021-01-01T00:00:00Z"`` or a datetime object.
        as_dict : bool, optional
            If True, the data will be returned as a list of dictionaries.
            Default is False.
        as_star_schema : bool, optional
            If True, the data will be returned in a star schema format.
            Default is False.
        outer_join_on_timestamp : bool
            If False, the data will be returned as a block-diagonal DataFrame,
            and it will contain the site and location columns. Besides,
            the timestamp column will not contain unique values since it will
            be repeated for each site and location. If False, the data will be
            returned as a full DataFrame, it will not contain the site and
            location columns, and the timestamp column will contain unique
            values.
        headers : Optional[Union[Dict[str, str]]], optional
            The headers to pass to the request. If None, the default headers
            will be used as ``{"accept": "application/json"}``. Default is None.
        data : pd.DataFrame
            The DataFrame to update with the data signals. If None, a new
            DataFrame will be created. Default is None.
        timeout : int, optional
            The timeout for the request in seconds. Default is 3600.
        threads : int, optional
            The number of threads to use for the request. Default is the number
            of CPU cores. If None, it will be set to the number of CPU cores.

        Returns
        -------
        Union[pd.DataFrame, Dict[str, Dict[str, pd.DataFrame]]]
            - The DataFrame containing the data signals, or
            - A dictionary containing the data signals divided by location, or
            - A dictionary containing the data signals in star schema format.
        """
        threads = U.set_threads_nr(threads)
        if data is None:
            data = pd.DataFrame()
        # Clean the DataFrame
        data_ = pd.DataFrame()
        # -- Step 1: Build the query object from GetData
        query = S.GetData(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
            outer_join_on_timestamp=outer_join_on_timestamp,
            as_dict=as_dict,
            as_star_schema=as_star_schema,
        )
        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        # nl = "\n"
        # l_ = f"\033[30;1mQuery:\033[0;34m {query_str.replace(' and ', f'{nl}       and ')}\n"
        # logging.info(l_)

        self._selected_metrics = query.get_selected_metrics(
            self._metrics_overview
        )
        grouped_metrics = query.group_metrics(self._metrics_overview)

        data_frames = []
        import concurrent.futures

        # fmt: off
        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=threads, thread_name_prefix="24SEA"
            ) as executor:
                future_to_data = {executor.submit(U.fetch_data_sync,
                                                  f"{self.base_url}data/",
                                                  site, location,
                                                  query.start_timestamp,
                                                  query.end_timestamp,
                                                  query.headers, group,
                                                  self._auth, timeout):
                                                      (site, location)
                    for (site, location), group in grouped_metrics}
                for future in concurrent.futures.as_completed(future_to_data):
                    data_frames.append(pd.DataFrame(future.result()))
        except RuntimeError:
            for (site, location), group in grouped_metrics:
                data_frames.append(pd.DataFrame(U.fetch_data_sync(
                    f"{self.base_url}data/", site, location,
                    query.start_timestamp, query.end_timestamp, query.headers,
                    group, self._auth, timeout))
                )
        # fmt: on
        # data_frames.append(pd.DataFrame(r_.json()))

        # if outer_join_on_timestamp is True, lose the location and site columns
        # and join on timestamp
        if outer_join_on_timestamp:
            for i, df in enumerate(data_frames):
                if df.empty:
                    continue
                data_frames[i] = df.set_index("timestamp")
                # drop site and location
                if "site" in data_frames[i].columns:
                    data_frames[i].drop(["site"], axis=1, inplace=True)
                if "location" in data_frames[i].columns:
                    data_frames[i].drop(["location"], axis=1, inplace=True)
            data_ = pd.concat([data_] + data_frames, axis=1, join="outer")
        else:
            data_ = pd.concat([data_] + data_frames, ignore_index=True)

        logging.info("\033[32;1m✔️ Data successfully retrieved.\033[0m")
        # fmt: off
        data.drop(data.index, inplace=True)
        data[data_.columns] = data_

        if as_dict:
            if as_star_schema:
                logging.info("\033[32;1m\033[32;1m\n⏳ Converting queried data "
                    "to \033[30;1mstar schema\033[0m...")
                return to_star_schema(data, self.selected_metrics(data) \
                                            .reset_index(names=["metric"]),
                                      as_dict=True,)
            return data.reset_index().to_dict("records")
        if as_star_schema:
            logging.info("\033[32;1m\033[32;1m\n⏳ Converting queried data to "
                         "\033[30;1mstar schema\033[0m...")
            return to_star_schema(data, self.selected_metrics(data) \
                                        .reset_index(names=["metric"]))
        # fmt: on
        return U.parse_timestamp(data) if not data.empty else data

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_stats(
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
        threads: Optional[int] = None,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """Get the metrics statistics (MAX, MIN, AVG) for the specified time
        range.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.
        threads: Optional[int]
            The number of threads to use for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        threads = U.set_threads_nr(threads)
        # Clean the DataFrame
        # -- Step 1: Build the query object from GetData
        query = S.GetStats(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        self._selected_metrics = query.get_selected_metrics(
            self._metrics_overview
        )
        grouped_metrics = query.group_metrics(self._metrics_overview)

        stats_list = []
        import concurrent.futures

        # fmt: off
        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=threads, thread_name_prefix="24SEA"
            ) as executor:
                future_to_data = {executor.submit(U.fetch_data_sync,
                                                  f"{self.base_url}stats/",
                                                  site, location,
                                                  query.start_timestamp,
                                                  query.end_timestamp,
                                                  query.headers, group,
                                                  self._auth, timeout):
                                                      (site, location)
                    for (site, location), group in grouped_metrics}
                for future in concurrent.futures.as_completed(future_to_data):
                    stats_list.append(future.result())
        except RuntimeError:
            for (site, location), group in grouped_metrics:
                stats_list.append(U.fetch_data_sync(f"{self.base_url}stats/",
                                                    site, location,
                                                    query.start_timestamp,
                                                    query.end_timestamp,
                                                    query.headers, group,
                                                    self._auth, timeout))
        # fmt: on
        # Parse stats list into a tidy table
        if not as_dict:
            return U.parse_stats_list(stats_list).pipe(
                U.get_stats_overview_info, self._selected_metrics
            )
        return (
            U.parse_stats_list(stats_list)
            .pipe(U.get_stats_overview_info, self._selected_metrics)
            .pipe(U.get_stats_as_dict)
        )

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_null_timestamps(
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
        threads: Optional[int] = None,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """Get the list of timestamps which the selected metrics have null
        values in the specified time range.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.
        threads: Optional[int]
            The number of threads to use for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        threads = U.set_threads_nr(threads)
        # Clean the DataFrame
        # -- Step 1: Build the query object from GetData
        query = S.GetStats(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        self._selected_metrics = query.get_selected_metrics(
            self._metrics_overview
        )
        grouped_metrics = query.group_metrics(self._metrics_overview)

        stats_list = []
        import concurrent.futures

        # fmt: off
        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=threads, thread_name_prefix="24SEA"
            ) as executor:
                future_to_data = {executor.submit(U.fetch_data_sync,
                                                  f"{self.base_url}null_timestamps/",
                                                  site, location,
                                                  query.start_timestamp,
                                                  query.end_timestamp,
                                                  query.headers, group,
                                                  self._auth, timeout):
                                                      (site, location)
                    for (site, location), group in grouped_metrics}
                for future in concurrent.futures.as_completed(future_to_data):
                    stats_list.append(future.result())
        except RuntimeError:
            for (site, location), group in grouped_metrics:
                stats_list.append(U.fetch_data_sync(f"{self.base_url}null_timestamps/",  # pylint: disable=C0301  # noqa:E501
                                                    site, location,
                                                    query.start_timestamp,
                                                    query.end_timestamp,
                                                    query.headers, group,
                                                    self._auth, timeout))
        # fmt: on
        # Parse stats list into a tidy table
        if not as_dict:
            return U.parse_stats_list(stats_list).pipe(
                U.get_stats_overview_info, self._selected_metrics
            )
        return (
            U.parse_stats_list(stats_list)
            .pipe(U.get_stats_overview_info, self._selected_metrics)
            .pipe(U.get_stats_as_dict)
        )

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_availability(
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        granularity: Union[str, int],
        sampling_interval_seconds: Optional[int] = None,
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
        threads: Optional[int] = None,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """Get the metrics statistics (MAX, MIN, AVG) for the specified time
        range.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        granularity: Union[str, int]
            The granularity of the data, can be a string, or an integer
            number of seconds. String values are restricted to
            "day", "week", "calendarmonth", "30days", or "365days". If
            "calendarmonth" is used, the availability will refer to the
            specific calendar month (e.g. January 2023), and not to a
            rolling period of 30 days.
        sampling_interval_seconds: Optional[int]
            The sampling interval in seconds. If None, the default value is
            used, which is 600 seconds (10 minutes).
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.
        threads: Optional[int]
            The number of threads to use for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        threads = U.set_threads_nr(threads)
        # Clean the DataFrame
        # -- Step 1: Build the query object from GetData
        query = S.GetAvailability(
            sites=sites,
            locations=locations,
            metrics=metrics,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            granularity=(
                granularity if granularity != "calendarmonth" else "day"
            ),
            sampling_interval_seconds=sampling_interval_seconds,
            as_dict=as_dict,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        self._selected_metrics = query.get_selected_metrics(
            self._metrics_overview
        )
        grouped_metrics = query.group_metrics(self._metrics_overview)

        av_list = []
        import concurrent.futures

        # fmt: off
        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=threads, thread_name_prefix="24SEA"
            ) as executor:
                future_to_data = {executor.submit(U.fetch_availability_sync,
                                                  f"{self.base_url}availability/",
                                                  site, location,
                                                  query.start_timestamp,
                                                  query.end_timestamp,
                                                  query.granularity,
                                                  query.sampling_interval_seconds,  # type: ignore  # pylint: disable=C0301  # noqa:E501
                                                  query.headers, group,
                                                  self._auth, timeout):
                                                      (site, location)
                    for (site, location), group in grouped_metrics}
                for future in concurrent.futures.as_completed(future_to_data):
                    av_list.append(future.result())
        except RuntimeError:
            for (site, location), group in grouped_metrics:
                    av_list.extend(U.fetch_data_sync(f"{self.base_url}availability/",
                                                     site, location,
                                                     query.start_timestamp,
                                                     query.end_timestamp,
                                                     query.headers, group,
                                                     self._auth, timeout))
        # fmt: on
        # av_list is a list of lists, I need to expand it to a single list
        df = pd.DataFrame(list(itertools.chain.from_iterable(av_list)))
        # Combine rows with the same timestamp
        df = df.groupby("bucket_start", as_index=True).first()
        df.index.name = "timestamp"
        df = U.parse_timestamp(df)
        if granularity == "calendarmonth":
            df = U.calendar_monthly_availability(
                df,
                start_timestamp=query.start_timestamp,
                end_timestamp=query.end_timestamp,
                sampling_interval_seconds=query.sampling_interval_seconds,  # type: ignore[arg-type]
            )
            # Ensure months with no data are represented (as 0 after fillna)
            if not df.empty and isinstance(df.index, pd.DatetimeIndex):
                start_dt = pd.to_datetime(query.start_timestamp)
                end_dt = pd.to_datetime(query.end_timestamp)
                tz = df.index.tz
                if tz is not None:
                    start_dt = (
                        start_dt.tz_convert(tz)
                        if start_dt.tzinfo is not None
                        else start_dt.tz_localize(tz)
                    )
                    end_dt = (
                        end_dt.tz_convert(tz)
                        if end_dt.tzinfo is not None
                        else end_dt.tz_localize(tz)
                    )
                end_adj = end_dt - pd.Timedelta(nanoseconds=1)
                start_month = pd.Timestamp(
                    year=start_dt.year,
                    month=start_dt.month,
                    day=1,
                    tz=tz,
                )
                end_month = pd.Timestamp(
                    year=end_adj.year,
                    month=end_adj.month,
                    day=1,
                    tz=tz,
                )
                full_index = pd.date_range(
                    start=start_month,
                    end=end_month,
                    freq="MS",
                    tz=tz,
                )
                df = df.reindex(full_index)
                df.index.name = "timestamp"
        # Convert NaN to 0
        df = df.fillna(0)
        if as_dict:
            return U.get_metrics_data_df_as_dict(
                df, selected_metrics=self._selected_metrics
            )
        return U.parse_timestamp(df)

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_oldest_timestamp(
        self,
        sites: Union[str, List[str]],
        locations: Optional[Union[List[str], str]],
    ) -> pd.DataFrame:
        """Get oldest timestamp for one or multiple sites (sync)."""
        all_dfs = []
        if isinstance(sites, str):
            sites = [sites]
        import concurrent.futures

        def fetch_oldest_timestamp_for_site(site):
            query = S.GetOldestTimestampSchema(site=site, locations=locations)
            query_df = query.get_selected_locations(self._metrics_overview)
            query_locs = query_df.location.to_list()
            query_site = query_df.site.iloc[0]
            # fmt: off
            params: Dict[str, Union[str, List[str]]] = {"project": query_site}
            params["locations"] = ",".join(query_locs).lower()
            r_ = U.handle_request(f"{self.base_url}oldest_timestamp/",
                                  params, self._auth,
                                  query.headers or {"accept": "application/json"})  # pylint: disable=C0301  # noqa: E501
            # fmt: on
            js = r_.json()
            if js == []:
                logging.warning(
                    f"\033[33;1m⚠️ No data found for {query.site}.\033[0m"
                )
            return pd.DataFrame(js)

        try:
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=len(sites), thread_name_prefix="24SEA"
            ) as executor:
                future_to_site = {
                    executor.submit(fetch_oldest_timestamp_for_site, site): site
                    for site in sites
                }
            for future in concurrent.futures.as_completed(future_to_site):
                all_dfs.append(future.result())
        except RuntimeError:
            for site in sites:
                all_dfs.append(fetch_oldest_timestamp_for_site(site))
        return (
            pd.concat(all_dfs, ignore_index=True) if all_dfs else pd.DataFrame()
        )

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def get_stats_predefined_intervals(
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
        threads: Optional[int] = None,
    ) -> Dict[str, Any]:
        """
        Run get_stats for predefined intervals:
          - all_time: (datetime.min -> datetime.max)
          - last_year: (now-365d -> now)
          - last_month: (now-30d -> now)
        """
        # fmt: off
        intervals = {
           "all_time": (datetime.datetime(2000, 1, 1, 0, 0, 0,
                                           tzinfo=datetime.timezone.utc),
                         datetime.datetime(3000, 1, 1, 0, 0, 0,
                                           tzinfo=datetime.timezone.utc)),
            "last_year": ("now-1Y", "now"),
            "last_month": ("now-1M", "now"),
        }
        # fmt: on
        results: Dict[str, Any] = {}
        for name, (start_ts, end_ts) in intervals.items():
            logging.info(
                f"\033[34;1mExecuting get_stats for interval: {name}\033[0m"
            )
            results[name] = self.get_stats(  # type: ignore
                sites=sites,
                locations=locations,
                metrics=metrics,
                start_timestamp=start_ts,
                end_timestamp=end_ts,
                as_dict=as_dict,
                headers=headers,
                timeout=timeout,
                threads=threads,
            )
        return results


class AsyncAPI(API):
    """Async version of the API class. Get data from 24sea API /datasignals
    asyncronously"""

    def __init__(self):
        super().__init__()

    @U.require_auth
    async def get_metrics_overview(self) -> Optional[pd.DataFrame]:
        """Asynchronously get metrics overview, authenticating if needed"""
        if not self.authenticated:
            self._lazy_authenticate()
        return self._metrics_overview

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_metrics(  # type: ignore[override]
        self,
        site: Optional[str] = None,
        locations: Optional[Union[str, List[str]]] = None,
        metrics: Optional[Union[str, List[str]]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        """
        Get the metrics names for a site asynchronously.
        """
        url = f"{self.base_url}metrics/"
        headers = headers or {"accept": "application/json"}
        if isinstance(locations, list):
            locations = ",".join(locations)
        if isinstance(metrics, list):
            metrics = ",".join(metrics)
        params = {"project": site, "locations": locations, "metrics": metrics}
        r_ = await U.handle_request_async(url, params, self._auth, headers)
        return r_.json()

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_data(  # type: ignore[override]
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        as_star_schema: bool = False,
        outer_join_on_timestamp: bool = True,
        headers: Optional[Dict[str, str]] = None,
        data: Optional[pd.DataFrame] = None,
        max_retries: int = 5,
        timeout: int = 1800,
    ) -> Optional[
        Union[
            pd.DataFrame,
            Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]],
            List[Union[Any, str]],
        ]
    ]:
        """
        Get the data signals from the 24SEA API asynchronously.
        """
        if data is None:
            data = pd.DataFrame()
        data_ = pd.DataFrame()
        query = S.GetData(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
            outer_join_on_timestamp=outer_join_on_timestamp,
            as_dict=as_dict,
            as_star_schema=as_star_schema,
        )

        metrics_overview = await self.get_metrics_overview()

        if metrics_overview is None:
            raise E.ProfileError(
                "\033[31mThe metrics overview is empty. Please authenticate "
                "first with the \033[1mauthenticate\033[22m method."
            )
        self._selected_metrics = query.get_selected_metrics(metrics_overview)
        grouped_metrics = query.group_metrics(metrics_overview)

        # Split tasks into chunks of 5 to avoid firing tens of requests together

        # fmt: off
        tasks = [U.fetch_data_async(f"{self.base_url}data/", site, location,
                                    query.start_timestamp, query.end_timestamp,
                                    query.headers, group, self._auth, timeout,
                                    max_retries)
                 for (site, location), group in grouped_metrics]
        chunk_size_dict = U.estimate_chunk_size(
            tasks,
            query.start_timestamp,   # type: ignore
            query.end_timestamp,     # type: ignore
            grouped_metrics,
            self._selected_metrics
        )
        # fmt: on
        data_frames = []
        data_frames = await U.gather_in_chunks(
            tasks, chunk_size=chunk_size_dict["chunk_size"], timeout=timeout  # type: ignore  # pylint: disable=C0301  # noqa:E501
        )

        if outer_join_on_timestamp:
            for i, df in enumerate(data_frames):
                if df.empty:
                    continue
                df = df.set_index("timestamp")
                for col in ["site", "location"]:
                    if col in df.columns:
                        df.drop(col, axis=1, inplace=True)
                data_frames[i] = df
            data_ = pd.concat([data_] + data_frames, axis=1, join="outer")
        else:
            data_ = pd.concat([data_] + data_frames, ignore_index=True)

        if all(
            getattr(r, "status_code", 200) == 200
            for r in data_frames
            if hasattr(r, "status_code")
        ):
            logging.info("\033[32;1m✔️ Data successfully retrieved.\033[0m")
        else:
            # If any response is not 200, return the response text(s)
            return [
                getattr(r, "text", "")
                for r in data_frames
                if hasattr(r, "status_code") and r.status_code != 200
            ]
        data.drop(data.index, inplace=True)
        for col in data_.columns:
            if col in data.columns:
                del data[col]
            data[col] = data_[col]
        if as_dict:
            if as_star_schema:
                logging.info(
                    "\033[32;1m\n⏳ Converting queried data to \033[30;1mstar schema\033[0m..."
                )
                return to_star_schema(
                    data,
                    self._selected_metrics.reset_index(drop=True),
                    as_dict=True,
                )
            return data.reset_index().to_dict("records")
        if as_star_schema:
            logging.info(
                "\033[32;1m\n⏳ Converting queried data to \033[30;1mstar schema\033[0m..."
            )
            return to_star_schema(
                data, self._selected_metrics.reset_index(drop=True)
            )
        return U.parse_timestamp(data) if not data.empty else data

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_oldest_timestamp(  # type: ignore
        self,
        sites: Union[str, List[str]],
        locations: Optional[Union[List[str], str]],
    ) -> pd.DataFrame:
        """Get oldest timestamp for one or multiple sites (async).

        Parameters
        ----------
        site: Union[str, List[str]]
            The site(s) to retrieve the oldest timestamp for.
        locations: Optional[Union[List[str], str]],
            The location(s) to retrieve the oldest timestamp for.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the oldest timestamp for the specified site(s)
            and location(s).
        """
        url = f"{self.base_url}oldest_timestamp/"
        tasks = []
        for site in sites:
            query = S.GetOldestTimestampSchema(site=site, locations=locations)
            query_df = query.get_selected_locations(self._metrics_overview)
            query_locs = query_df.location.to_list()
            query_site = query_df.site.iloc[0]
            # fmt: off
            tasks.append(U.fetch_oldest_timestamp_async(
                    url,
                    query_site,
                    ",".join(query_locs),
                    query.headers,  # type: ignore
                    self._auth,
                    3600,
                    5,
                    False,
                )
            )
        results = await asyncio.gather(*tasks)
        if not results:
            return pd.DataFrame()
        return pd.concat(results, ignore_index=True)

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_stats(  # type: ignore
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """
        Get the metrics statistics (MAX, MIN, AVG) for the specified time
        range asynchronously.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        # Clean the DataFrame
        # -- Step 1: Build the query object from GetStats
        query = S.GetStats(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        metrics_overview = await self.get_metrics_overview()

        if metrics_overview is None:
            raise E.ProfileError(
                "\033[31mThe metrics overview is empty. Please authenticate "
                "first with the \033[1mauthenticate\033[22m method."
            )

        self._selected_metrics = query.get_selected_metrics(metrics_overview)
        grouped_metrics = query.group_metrics(metrics_overview)

        # fmt: off
        tasks = [U.fetch_data_async(f"{self.base_url}stats/", site, location,
                                    query.start_timestamp, query.end_timestamp,
                                    query.headers, group, self._auth, timeout,
                                    5, True)
                 for (site, location), group in grouped_metrics]
        # fmt: on
        stats_list = await U.gather_in_chunks(
            tasks, chunk_size=1000, timeout=timeout  # type: ignore  # pylint: disable=C0301  # noqa:E501
        )

        # Parse stats list into a tidy table
        if not as_dict:
            return U.parse_stats_list(stats_list).pipe(
                U.get_stats_overview_info, self._selected_metrics
            )
        return (
            U.parse_stats_list(stats_list)
            .pipe(U.get_stats_overview_info, self._selected_metrics)
            .pipe(U.get_stats_as_dict)
        )

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_null_timestamps(  # type: ignore
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """
        Get the metrics statistics (MAX, MIN, AVG) for the specified time
        range asynchronously.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        # Clean the DataFrame
        # -- Step 1: Build the query object from GetStats
        query = S.GetStats(
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            sites=sites,
            locations=locations,
            metrics=metrics,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        metrics_overview = await self.get_metrics_overview()

        if metrics_overview is None:
            raise E.ProfileError(
                "\033[31mThe metrics overview is empty. Please authenticate "
                "first with the \033[1mauthenticate\033[22m method."
            )

        self._selected_metrics = query.get_selected_metrics(metrics_overview)
        grouped_metrics = query.group_metrics(metrics_overview)

        # fmt: off
        tasks = [U.fetch_data_async(f"{self.base_url}null_timestamps/",
                                    site, location,
                                    query.start_timestamp, query.end_timestamp,
                                    query.headers, group, self._auth, timeout,
                                    5, True)
                 for (site, location), group in grouped_metrics]
        # fmt: on
        stats_list = await U.gather_in_chunks(
            tasks, chunk_size=1000, timeout=timeout  # type: ignore  # pylint: disable=C0301  # noqa:E501
        )

        # Parse stats list into a tidy table
        if not as_dict:
            return U.parse_stats_list(stats_list).pipe(
                U.get_stats_overview_info, self._selected_metrics
            )
        return (
            U.parse_stats_list(stats_list)
            .pipe(U.get_stats_overview_info, self._selected_metrics)
            .pipe(U.get_stats_as_dict)
        )

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_availability(  # type: ignore
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        start_timestamp: Union[str, datetime.datetime],
        end_timestamp: Union[str, datetime.datetime],
        granularity: Union[str, int],
        sampling_interval_seconds: Optional[int] = None,
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
    ) -> Union[
        pd.DataFrame, Dict[str, Union[Dict[str, pd.DataFrame], Dict[str, Any]]]
    ]:
        """Get the metrics statistics (MAX, MIN, AVG) for the specified time
        range.

        Parameters
        ----------
        sites: Optional[Union[List, str]]
            The sites to filter the data.
        locations: Optional[Union[List, str]]
            The locations to filter the data.
        metrics: Union[List, str]
            The metrics to retrieve.
        start_timestamp: Union[str, datetime.datetime]
            The start timestamp for the data retrieval.
        end_timestamp: Union[str, datetime.datetime]
            The end timestamp for the data retrieval.
        granularity: Union[str, int]
            The granularity of the data, can be a string, or an integer
            number of seconds. String values are restricted to
            "day", "week", "calendarmonth", "30days", or "365days". If
            "calendarmonth" is used, the availability will refer to the
            specific calendar month (e.g. January 2023), and not to a
            rolling period of 30 days.
        sampling_interval_seconds: Optional[int]
            The sampling interval in seconds. If None, the default value is
            used, which is 600 seconds (10 minutes).
        as_dict: bool
            Whether to return the data as a dictionary.
        headers: Optional[Union[Dict[str, str]]]
            Headers to include in the request.
        timeout: int
            The timeout for the request.
        threads: Optional[int]
            The number of threads to use for the request.

        Returns
        -------
        Union[DataFrame, Dict[str, Union[Dict[str, DataFrame], Dict[str, Any]]]]
            The retrieved data.
        """
        # Build query
        query = S.GetAvailability(
            sites=sites,
            locations=locations,
            metrics=metrics,
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            granularity=(
                granularity if granularity != "calendarmonth" else "day"
            ),
            sampling_interval_seconds=sampling_interval_seconds,
            as_dict=as_dict,
            headers=headers,
        )

        logging.info(
            f"\n\033[32;1mRequested time range:\033[0;34m "
            f"From {str(query.start_timestamp)[:-4].replace('T', ' ')} UTC "
            f"To {str(query.end_timestamp)[:-4].replace('T', ' ')} UTC\n\033[0m"
        )

        metrics_overview = await self.get_metrics_overview()
        if metrics_overview is None:
            raise E.ProfileError(
                "\033[31mThe metrics overview is empty. Please authenticate "
                "first with the \033[1mauthenticate\033[22m method."
            )

        self._selected_metrics = query.get_selected_metrics(metrics_overview)
        grouped_metrics = query.group_metrics(metrics_overview)

        # Build async tasks
        tasks = [
            U.fetch_availability_async(
                f"{self.base_url}availability/",
                site,
                location,
                query.start_timestamp,
                query.end_timestamp,
                query.granularity,
                query.sampling_interval_seconds,  # type: ignore
                query.headers,
                group,
                self._auth,
                timeout,
                5,  # max_retries
            )
            for (site, location), group in grouped_metrics
        ]

        # Estimate chunk size (reuse estimator; good enough for availability)
        chunk_info = U.estimate_chunk_size(
            tasks,
            query.start_timestamp,  # type: ignore
            query.end_timestamp,  # type: ignore
            grouped_metrics,
            self._selected_metrics,
        )
        av_list = await U.gather_in_chunks(
            tasks, chunk_size=chunk_info["chunk_size"], timeout=timeout  # type: ignore  # pylint: disable=C0301  # noqa:E501
        )

        # Flatten (each task returns a list[dict])
        flat = list(itertools.chain.from_iterable(av_list))
        if not flat:
            return pd.DataFrame()
        df = pd.DataFrame(flat)
        if df.empty:
            return df
        if "bucket_start" not in df.columns:
            # fmt: off
            logging.warning("\033[33;1mAvailability response missing "
                            "'bucket_start' column.\033[0m")
            # fmt: on
            return df
        df = df.groupby("bucket_start", as_index=True).first()
        df.index.name = "timestamp"
        if granularity == "calendarmonth":
            df = U.calendar_monthly_availability(
                df,
                start_timestamp=query.start_timestamp,
                end_timestamp=query.end_timestamp,
                sampling_interval_seconds=query.sampling_interval_seconds,  # type: ignore[arg-type]
            )
            # Ensure months with no data are represented (as 0 after fillna)
            if not df.empty and isinstance(df.index, pd.DatetimeIndex):
                start_dt = pd.to_datetime(query.start_timestamp)
                end_dt = pd.to_datetime(query.end_timestamp)
                tz = df.index.tz
                if tz is not None:
                    start_dt = (
                        start_dt.tz_convert(tz)
                        if start_dt.tzinfo is not None
                        else start_dt.tz_localize(tz)
                    )
                    end_dt = (
                        end_dt.tz_convert(tz)
                        if end_dt.tzinfo is not None
                        else end_dt.tz_localize(tz)
                    )
                end_adj = end_dt - pd.Timedelta(nanoseconds=1)
                start_month = pd.Timestamp(
                    year=start_dt.year,
                    month=start_dt.month,
                    day=1,
                    tz=tz,
                )
                end_month = pd.Timestamp(
                    year=end_adj.year,
                    month=end_adj.month,
                    day=1,
                    tz=tz,
                )
                full_index = pd.date_range(
                    start=start_month,
                    end=end_month,
                    freq="MS",
                    tz=tz,
                )
                df = df.reindex(full_index)
                df.index.name = "timestamp"
        # Convert NaN to 0
        df = df.fillna(0)
        if as_dict:
            return U.get_metrics_data_df_as_dict(
                df, selected_metrics=self._selected_metrics
            )
        return U.parse_timestamp(df)

    @U.require_auth
    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    async def get_stats_predefined_intervals(  # type: ignore
        self,
        sites: Optional[Union[List, str]],
        locations: Optional[Union[List, str]],
        metrics: Union[List, str],
        as_dict: bool = False,
        headers: Optional[Dict[str, str]] = None,
        timeout: int = 3600,
    ) -> Dict[str, Any]:
        """
        Async version of get_stats_predefined_intervals.
        """

        # fmt: off
        intervals = {
           "all_time": (datetime.datetime(2000, 1, 1, 0, 0, 0,
                                           tzinfo=datetime.timezone.utc),
                         datetime.datetime(3000, 1, 1, 0, 0, 0,
                                           tzinfo=datetime.timezone.utc)),
            "last_year": ("now-1Y", "now"),
            "last_month": ("now-1M", "now"),
        }
        # fmt: on
        async def _run(start_ts, end_ts):
            return await self.get_stats(  # type: ignore
                sites=sites,
                locations=locations,
                metrics=metrics,
                start_timestamp=start_ts,
                end_timestamp=end_ts,
                as_dict=as_dict,
                headers=headers,
                timeout=timeout,
            )

        tasks = {name: _run(s, e) for name, (s, e) in intervals.items()}
        gathered = await asyncio.gather(*tasks.values())
        return dict(zip(tasks.keys(), gathered))


def to_category_value(
    data: Union[pd.DataFrame, Dict[str, Dict[str, pd.DataFrame]]],
    metrics_overview: pd.DataFrame,
    keep_stat_in_metric_name: bool = False,
) -> pd.DataFrame:
    """
    Categorize the data based on the metrics overview.

    Parameters
    ----------
    data : Union[pd.DataFrame, Dict[str, Dict[str, pd.DataFrame]]]
        The data to be categorized. It can be either a DataFrame or a dictionary
        of DataFrames.
    metrics_overview : pd.DataFrame
        A DataFrame containing the information about the metrics.
    keep_stat_in_metric_name : bool, optional
        Whether to keep the statistic in the metric name, by default True.

    Returns
    -------
    Union[pd.DataFrame, Dict[str, Dict[str, pd.DataFrame]]]
        The data in category-value format, based on the metrics overview.

    Notes
    -----
    The function performs the following steps:
    1. Transforms the data dictionary into a DataFrame if necessary.
    2. Resets the index and converts the timestamp column to datetime.
    3. Melts the data to long format.
    4. Merges the melted data with the metrics overview DataFrame.
    5. Renames columns for consistency.
    6. Extracts latitude and heading information from the metric names.
    7. Extracts sub-assembly information from the metric names.
    8. Reorders the columns.
    9. Optionally appends the statistic to the metric name.
    10. Drops the rows where the metric name is "index", "site" or "location".

    Example
    -------
    >>> import pandas as pd
    >>> from typing import Union, Dict
    >>> data = {
    ...     'timestamp': ['2021-01-01', '2021-01-02'],
    ...     'mean_WF_A01_TP_SG_LAT005_DEG000': [1.0, 1.1],
    ...     'mean_WF_A02_TP_SG_LAT005_DEG000': [2.0, 2.1]
    ... }
    >>> metrics_overview = pd.DataFrame({
    ...     'metric': ['mean_WF_A01_TP_SG_LAT005_DEG000',
    ...                'mean_WF_A02_TP_SG_LAT005_DEG000'],
    ...     'short_hand': ['TP_SG_LAT005_DEG000', 'TP_SG_LAT005_DEG000'],
    ...     'statistic': ['mean', 'mean'],
    ...     'unit': ['unit', 'unit'],
    ...     'site': ['WindFarm', 'WindFarm'],
    ...     'location': ['WFA01', 'WFA02'],
    ...     'data_group': ['SG', 'SG'],
    ...     'site_id': ['WF', 'WF'],
    ...     'location_id': ['A01', 'A02']
    ... })
    >>> categorized = to_category_value(data, metrics_overview)
    >>> categorized
    +------------+--------------------------------+-------+------+-----------+---------------------+---------+-------------+-----+---------+-----------+----------+--------------+
    | timestamp  | full_metric_name               | value | unit | statistic | short_hand          | site_id | location_id | lat | heading | site      | location | metric_group |
    +============+================================+=======+======+===========+=====================+=========+=============+=====+=========+===========+==========+==============+
    | 2021-01-01 | mean_WF_A01_TP_SG_LAT005_DEG000| 1.0   | unit | mean      | TP_SG_LAT005_DEG000 | WF      | A01         | 5.0 | 0.0     | WindFarm  | WFA01    | SG           |
    +------------+--------------------------------+-------+------+-----------+---------------------+---------+-------------+-----+---------+-----------+----------+--------------+
    | 2021-01-02 | mean_WF_A01_TP_SG_LAT005_DEG000| 1.1   | unit | mean      | TP_SG_LAT005_DEG000 | WF      | A01         | 5.0 | 0.0     | WindFarm  | WFA01    | SG           |
    +------------+--------------------------------+-------+------+-----------+---------------------+---------+-------------+-----+---------+-----------+----------+--------------+
    | 2021-01-01 | mean_WF_A02_TP_SG_LAT005_DEG000| 2.0   | unit | mean      | TP_SG_LAT005_DEG000 | WF      | A02         | 5.0 | 0.0     | WindFarm  | WFA02    | SG           |
    +------------+--------------------------------+-------+------+-----------+---------------------+---------+-------------+-----+---------+-----------+----------+--------------+
    | 2021-01-02 | mean_WF_A02_TP_SG_LAT005_DEG000| 2.1   | unit | mean      | TP_SG_LAT005_DEG000 | WF      | A02         | 5.0 | 0.0     | WindFarm  | WFA02    | SG           |
    +------------+--------------------------------+-------+------+-----------+---------------------+---------+-------------+-----+---------+-----------+----------+--------------+
    """
    if isinstance(data, dict):
        data = pd.DataFrame(data)
    # Categorize the data
    data = U.parse_timestamp(data, keep_index_only=False).reset_index(drop=True)
    # Melt the data
    categorized = data.melt(
        id_vars=["timestamp"], var_name="metric", value_name="value"
    )
    # Merge the melted data with the metrics overview DataFrame
    categorized = categorized.merge(
        metrics_overview, how="left", left_on="metric", right_on="metric"
    )
    # Rename the columns
    categorized.rename(
        columns={"unit_str": "unit", "data_group": "metric_group"}, inplace=True
    )
    # Get the lat, and heading from the metric name
    categorized["lat"] = (
        categorized["metric"].str.extract(r"(_LAT)(\w{3})")[1].astype(float)
    )
    categorized["heading"] = (
        categorized["metric"].str.extract(r"(_DEG)(\w{3})")[1].astype(float)
    )
    # Now get the subassembly from the metric name.
    try:
        if pd_version > "2.0.0":
            categorized["sub_assembly"] = (
                categorized["metric"]
                .str.extract(r"(_TP_)|(_TW_)|(_MP_)")
                .bfill(axis=1)
                .infer_objects(copy=False)[0]  # type: ignore
                .str.replace("_", "")
            )
        else:
            raise ImportError
    except ImportError:
        categorized["metric"] = pd.Series(categorized["metric"], dtype="string")
        categorized["sub_assembly"] = (
            categorized["metric"]
            .str.extract(r"(_TP_)|(_TW_)|(_MP_)")
            .bfill(axis=1)
            .apply(lambda x: x[0], axis=1)
            .str.replace("_", "")
        )
    # Reorder the columns
    # fmt: off
    columns = ["timestamp", "metric", "value", "unit", "statistic",
               "short_hand", "site_id", "location_id", "sub_assembly", "lat",
               "heading", "site", "location", "metric_group"]
    # fmt: on
    categorized = categorized[columns]
    if keep_stat_in_metric_name:
        categorized["stat_short_hand"] = (
            categorized["statistic"] + "_" + categorized["short_hand"]
        )
    # Drop the rows where the value of column metric is "index", "site" or
    # "location"
    # fmt: on
    categorized = categorized[
        ~categorized["metric"].isin(["index", "site", "location"])
    ]
    return categorized.reset_index(drop=True)


def to_star_schema(
    data: Union[pd.DataFrame, Dict[str, List[Dict[str, Any]]]],
    metrics_overview: Optional[pd.DataFrame] = None,
    as_dict: bool = False,
    convert_object_columns_to_string: bool = False,
    _username: Optional[str] = None,
    _password: Optional[str] = None,
) -> Optional[Union[Dict[str, Any], pd.DataFrame]]:
    """
    Transforms the data and metrics_overview into a star schema format for
    analytical purposes.

    Parameters
    ----------
    data : Union[pd.DataFrame, Dict[str, list[dict[str, Any]]]]
        A DataFrame or dictionary representing the raw data. The keys are column
        column names, and the values are lists of data.
        Must include a "timestamp" column or have indices that can be converted
        to timestamps.
    metrics_overview : pd.DataFrame
        A DataFrame containing metadata for metrics, including the following
        required columns:
        - | 'metric': The metric names (must match column names in `data`).
        - | 'short_hand': Short descriptive names for the metrics.
        - | 'description': Detailed descriptions of the metrics.
        - | 'statistic': Aggregation or statistical operation (e.g., mean,
        | std).
        - | 'unit_str': The units for the metrics.
        - | 'location': Location identifiers.
        - | 'site': Windfarm identifiers.
        - | 'data_group': Grouping of data (e.g., "scada").
    as_dict : bool, optional
        If True, the data will be returned as a dictionary. Default is False.
    convert_object_columns_to_string : bool, optional
        If True, convert object columns in the DataFrame to string. This feature
        is useful if importing the DataFrame within a database so that the
        'value' column can be stored as a float, since the non-float values
        will be stored as NULL. Default is False.
    _username : Optional[str]
        The username for authentication. If None, the username will be inferred
        from the environment variables.
    _password : Optional[str]
        The password for authentication. If None, the password will be inferred
        from the environment variables.

    Returns
    -------
    dict[str, pd.DataFrame]
        A dictionary containing the following tables:

        - | 'FactData': The fact table linking metrics to timestamps,
          | locations, metric IDs, and their values as columns.
        - | 'FactPivotData': The fact table in pivot format, i.e. containing
          | timestamp, location, and "statistic" + "short_hand" metric names
          | as columns. This pivoted format is the ones used generally by
          | BI tools and databases such as InfluxDB.
        - | 'DimMetric': Dimension table for metrics, including metric ID,
          | short name, description, statistic, and unit.
        - | 'DimWindFarm': Dimension table for wind farms, including
          | locations and sites.
        - | 'DimCalendar': Dimension table for time, including date parts
          | (year, month, day, hour, minute).
        - | 'DimDataGroup': Dimension table for data groups.

    Raises
    ------
    ValueError
        If required columns are missing in `data` or `metrics_overview`.
    KeyError
        If the `metric` column in `metrics_overview` contains values not present
        in `data`.

    Example
    -------
    >>> import pandas as pd
    >>> data = {
    ...     'timestamp': ['2020-01-01T00:00:00Z', '2020-01-01T00:10:00Z'],
    ...     'mean_WF_A01_winddirection': [257.445, 262.03],
    ...     'std_WF_A01_windspeed': [1.5165, 1.7966]
    ... }
    >>> metrics_overview = pd.DataFrame({
    ...     'metric': ['mean_WF_A01_winddirection', 'std_WF_A01_windspeed'],
    ...     'short_hand': ['winddirection', 'windspeed'],
    ...     'description': ['Wind direction', 'Wind speed'],
    ...     'statistic': ['mean', 'std'],
    ...     'unit_str': ['°', 'm/s'],
    ...     'location': ['WFA01', 'WFMA4'],
    ...     'site': ['windfarm', 'windfarm'],
    ...     'data_group': ['scada', 'scada']
    ... })
    >>> result = to_star_schema(data, metrics_overview)
    >>> for key, df in result.items():
    ...     print(f"{key}: {df.to_markdown()}")
    """
    # fmt: off
    # Input validation
    if metrics_overview is None:
        try:
            api = API()
            if _username is not None and _password is not None:
                api.authenticate(_username, _password)
            else:
                api._lazy_authenticate()
            metrics_overview = api._metrics_overview
        except AttributeError:
            raise ValueError("Failed to retrieve metrics overview from the "
                             "datasignals accessor. ``metrics_overview`` must "
                             "be provided as an argument.")
    if metrics_overview is None:
        raise ValueError("Failed to retrieve metrics overview from the "
                             "datasignals accessor. ``metrics_overview`` must "
                             "be provided as an argument.")
    req_metrics_cols = {"metric", "short_hand", "description", "statistic",
                        "unit_str", "location", "site", "data_group"}
    if not req_metrics_cols.issubset(metrics_overview.columns):
        raise ValueError("metrics_overview must contain the following columns: "
                         f"{req_metrics_cols}.\n Found: {metrics_overview.columns}")  # noqa: E501  # pylint: disable=C0301

    if isinstance(data, dict):
        data = pd.DataFrame(data)

    if data.empty or data is None:
        return data
    # If site and location are not present in the data, drop them
    data.drop(columns=['site', 'location'], errors='ignore', inplace=True)

    # Ensure timestamps are datetime
    data = U.parse_timestamp(data, keep_index_only=False).reset_index(drop=True)

    if "timestamp" not in data.columns:
        raise ValueError("`data` must include a 'timestamp' column or indices "
                         "convertible to timestamps.")

    missing_metrics = set(data.columns) \
                      - {"timestamp", "index", "site", "location"} \
                      - set(metrics_overview["metric"])
    if missing_metrics:
        raise KeyError("The following metrics in `data` are "
                      f"missing from `metrics_overview`: {missing_metrics}")
    # fmt: on

    # Reshape the data to long format
    fact_metrics = data.melt(
        id_vars=["timestamp"], var_name="metric", value_name="value"
    )

    # Create DimMetric with metric_id
    dim_metric = metrics_overview[
        ["metric", "short_hand", "description", "statistic", "unit_str"]
    ].drop_duplicates()
    dim_metric.rename(columns={"short_hand": "short_str"}, inplace=True)

    # Generate unique metric_id as a composite key
    dim_metric["metric_id"] = (
        dim_metric["statistic"] + "_" + dim_metric["short_str"]
    )

    # Map metric_id to the fact table
    metric_to_id = dim_metric.set_index("metric")["metric_id"]
    fact_metrics["metric_id"] = fact_metrics["metric"].map(metric_to_id)

    # Dimension table for WindFarm
    dim_windfarm = (
        metrics_overview[["location", "site"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )

    # Map site and location to fact table
    metric_to_site_location = metrics_overview[
        ["metric", "site", "location"]
    ].set_index("metric")
    fact_metrics["location"] = fact_metrics["metric"].map(
        metric_to_site_location["location"]
    )
    fact_metrics["site"] = fact_metrics["metric"].map(
        metric_to_site_location["site"]
    )

    # Dimension table for Calendar
    dim_calendar = (
        fact_metrics[["timestamp"]].drop_duplicates().reset_index(drop=True)
    )
    dim_calendar["year"] = dim_calendar["timestamp"].dt.year
    dim_calendar["month"] = dim_calendar["timestamp"].dt.month
    dim_calendar["day"] = dim_calendar["timestamp"].dt.day
    dim_calendar["hour"] = dim_calendar["timestamp"].dt.hour
    dim_calendar["minute"] = dim_calendar["timestamp"].dt.minute

    # Dimension table for DataGroup
    dim_data_group = (
        metrics_overview[["data_group"]]
        .drop_duplicates()
        .reset_index(drop=True)
    )

    # Add data_group_id to fact_metrics
    data_group_map = (
        metrics_overview[["metric", "data_group"]]
        .drop_duplicates()
        .set_index("metric")["data_group"]
    )
    fact_metrics["data_group"] = fact_metrics["metric"].map(data_group_map)
    # Select and reorder columns for the fact table
    fact_metrics = fact_metrics[
        ["timestamp", "location", "data_group", "metric_id", "value"]
    ]
    # Convert value column to float in fact_metrics
    if convert_object_columns_to_string:
        fact_metrics = U.column_to_type(fact_metrics, "value", float)
    # Add year and month columns to fact_metrics for partitioning
    fact_metrics["year"] = fact_metrics["timestamp"].dt.year
    fact_metrics["month"] = fact_metrics["timestamp"].dt.month
    fact_metrics["day"] = fact_metrics["timestamp"].dt.day
    # Create another fact table with metric_id as columns
    fact_pivot_metrics = (
        fact_metrics.pivot_table(
            index=["timestamp", "location"],
            columns="metric_id",
            values="value",
            aggfunc="first",
        )
        .reset_index()
        .rename_axis(None, axis=1)
    ).sort_values(
        ["location", "timestamp"], ascending=[True, True], ignore_index=True
    )
    # Convert object columns to string in fact_pivot_metrics
    if convert_object_columns_to_string:
        object_columns = fact_pivot_metrics.select_dtypes(
            include=["object"]
        ).columns  # noqa: E501  # pylint: disable=C0301
        for col in object_columns:
            fact_pivot_metrics = U.column_to_type(fact_pivot_metrics, col, str)

    # fmt: off
    if as_dict:
        return {"FactData": fact_metrics.to_dict("records"),
                "FactPivotData": fact_pivot_metrics \
                                 .drop_duplicates(subset=["location",
                                                          "timestamp"],
                                                  ignore_index=True,
                                                  keep="first") \
                                                  .to_dict("records"),
                "DimMetric": dim_metric[["metric_id", "short_str", "statistic",
                                         "description", "unit_str"]] \
                             .drop_duplicates(subset=["metric_id"], \
                                              keep="first").to_dict("records"),
                "DimWindFarm": dim_windfarm.to_dict("records"),
                "DimCalendar": dim_calendar.to_dict("records"),
                "DimDataGroup": dim_data_group.to_dict("records"),
        }
    # fmt: on
    return {
        "FactData": fact_metrics,
        "FactPivotData": fact_pivot_metrics.drop_duplicates(
            subset=["location", "timestamp"], ignore_index=True, keep="first"
        ),
        "DimMetric": dim_metric[
            ["metric_id", "short_str", "statistic", "description", "unit_str"]
        ].drop_duplicates(subset=["metric_id"], keep="first"),
        "DimWindFarm": dim_windfarm,
        "DimCalendar": dim_calendar,
        "DimDataGroup": dim_data_group,
    }



# -*- coding: utf-8 -*-
"""Data signals types."""
import datetime
import logging
from typing import Dict, List, Optional, Union

import pandas as pd
from pandas.core.groupby import DataFrameGroupBy
from pydantic import PositiveInt

from .. import exceptions as E
from .. import utils as U


# Shared base class for data signal schemas
class BaseDataSignalSchema(U.BaseModel):
    """
    Shared base schema for data signals requests.
    Includes common fields and validation logic for timestamp, sites, locations,
    metrics, and headers. Inherit from this class to create specific schemas for
    data retrieval and stats retrieval.
    """

    start_timestamp: Union[datetime.datetime, str]
    end_timestamp: Union[datetime.datetime, str]
    sites: Optional[Union[str, List[str]]] = None
    locations: Optional[Union[str, List[str]]] = None
    metrics: Union[str, List[str]]
    as_dict: Optional[bool] = False
    as_star_schema: Optional[bool] = False
    headers: Optional[Dict[str, str]] = {"accept": "application/json"}

    @U.field_validator("start_timestamp", "end_timestamp", mode="before")
    def validate_timestamp(cls, v: Union[datetime.datetime, str]) -> str:
        """Validate and format timestamp input."""
        if isinstance(v, str):
            try:
                datetime.datetime.strptime(v, "%Y-%m-%dT%H:%M:%SZ").astimezone(
                    datetime.timezone.utc
                )
            except Exception:
                try:
                    from shorthand_datetime import parse_shorthand_datetime

                    return parse_shorthand_datetime(v).strftime(  # type: ignore
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                except Exception as exc:
                    raise ValueError(
                        "\033[31mIncorrect start timestamp format, expected "
                        "one of the following formats:"
                        "\n               \033[1m• 'YYYY-MM-DDTHH:MM:SSZ'"
                        "\033[22m, \n               \033[1m• shorthand_datetime"
                        "-compatible string\033[22m "
                        "(https://pypi.org/project/shorthand-datetime/), or, "
                        "\n               "
                        "\033[1m• datetime.datetime\033[22m object.\033[0m\n\n"
                        "Exception originated from\n" + str(exc)
                    )
        if isinstance(v, datetime.datetime):
            return v.strftime("%Y-%m-%dT%H:%M:%SZ")
        return v

    @U.field_validator("sites", "locations", mode="before")
    def validate_sites_locations(cls, v):
        """Validate and format sites and locations input."""
        if isinstance(v, str):
            v = [v]
        if isinstance(v, List):
            v = [str(item).lower() for item in v]
        return v

    @U.field_validator("metrics", mode="before")
    def validate_metrics(cls, v):
        """Validate and format metrics input."""
        if isinstance(v, str):
            v = [v]
        if isinstance(v, List):
            # fmt: off
            v = [item.replace(" ", ".*")
                    #  .replace("_", ".*")
                     .replace("-", ".*") for item in v]
            # fmt: on
        return "|".join(v)

    @U.field_validator("headers", mode="before")
    def validate_headers(cls, v):
        """Validate and format headers input."""
        if v is None:
            return {"accept": "application/json"}
        return v

    @property
    def query_str(self) -> str:
        """Build a query string based on self."""
        return build_query_str(self)

    # @lru_cache
    def get_selected_metrics(
        self, df: Optional[pd.DataFrame], log: bool = True
    ) -> pd.DataFrame:
        """Calculate the selected metrics DataFrame based on a df (metrics
        overview) and the query object."""
        if df is None:
            raise ValueError("The provided DataFrame cannot be None.")
        return get_selected_metrics(self, df, log)

    # @lru_cache
    def group_metrics(self, df: Optional[pd.DataFrame]) -> DataFrameGroupBy:
        """Group metrics by site and location."""
        if df is None:
            raise ValueError("The provided DataFrame cannot be None.")
        s_m = self.get_selected_metrics(df, False)
        if self.metrics in (["all"], "all"):
            # For "all" metrics, create groups more efficiently
            sites_locs = s_m.assign(
                site=lambda x: x["site"].str.lower(),
                location=lambda x: x["location"].str.upper(),
            )[["site", "location"]].drop_duplicates()

            # Create grouped_metrics directly without intermediate steps
            return sites_locs.assign(metric="all").groupby(["site", "location"])
        else:
            # For specific metrics, group directly
            return s_m.assign(
                site=lambda x: x["site"].str.lower(),
                location=lambda x: x["location"].str.upper(),
            ).groupby(["site", "location"])


class GetOldestTimestampSchema(U.BaseModel):
    """
    Schema for oldest timestamp requests.
    """

    site: str  # renamed from 'sites' to match caller
    locations: Optional[Union[str, List[str]]] = None
    headers: Optional[Dict[str, str]] = {"accept": "application/json"}

    @U.field_validator("locations", mode="before")
    def validate_locations(cls, v) -> Optional[str]:
        """Normalize locations to a comma-separated lower-case string."""
        if v is None:
            return v
        if isinstance(v, list):
            return ",".join([str(item).lower() for item in v])
        if isinstance(v, str):
            return v.lower()
        raise TypeError("locations must be a string or list of strings")

    @U.field_validator("site", mode="before")
    def validate_site(cls, v):
        """Normalize site to a lower-case string."""
        if isinstance(v, str):
            return v.lower()
        raise TypeError("site must be a string")

    @U.field_validator("headers", mode="before")
    def validate_headers(cls, v) -> Dict[str, str]:
        if v is None:
            return {"accept": "application/json"}
        if isinstance(v, dict):
            return v
        raise TypeError("headers must be a dictionary")

    @property
    def query_str(self) -> str:
        # fmt: off
        if self.locations is None:
            return ("(site_id.str.lower() == @query.site or "
                    "site.str.lower() == @query.site)")
        return ("(site_id.str.lower() == @query.site or "
                "site.str.lower() == @query.site) "
                "and (location.str.lower() == @query.locations.split(',') or "
                "location_id.str.lower() == @query.locations.split(','))")
        # fmt: on

    def get_selected_locations(
        self, df: Optional[pd.DataFrame]
    ) -> pd.DataFrame:
        """Get the selected locations DataFrame based on the query object."""
        if df is None:
            raise ValueError("The provided DataFrame cannot be None.")
        return get_selected_locations(self, df)


class GetData(BaseDataSignalSchema):
    """
    A pydantic schema for the data signals (data retrieval).
    Includes all common fields plus 'outer_join_on_timestamp' for join behavior.
    """

    outer_join_on_timestamp: bool

    @U.field_validator("outer_join_on_timestamp", mode="before")
    def validate_outer_join_on_timestamp(cls, v):
        if v is None:
            return False
        return v


class GetStats(BaseDataSignalSchema):
    """
    A pydantic schema for the data signals (stats retrieval).
    Inherits all common fields from the base schema.
    """

    # Placeholder for future stats-specific fields or methods
    # Currently, it inherits all common fields from BaseDataSignalSchema
    pass


class GetAvailability(BaseDataSignalSchema):
    """
    A pydantic schema for the data signals (availability retrieval).
    Inherits all common fields from the base schema.
    """

    # Placeholder for future availability-specific fields or methods
    # Currently, it inherits all common fields from BaseDataSignalSchema
    granularity: Union[str, PositiveInt]
    sampling_interval_seconds: Optional[int] = None

    @U.field_validator("granularity", mode="before")
    def validate_granularity(cls, v):
        valid_granularities = ["day", "week", "30days", "365days"]
        if isinstance(v, str):
            if v not in valid_granularities:
                # fmt: off
                raise ValueError(f"Invalid granularity: {v}. "
                                 f"Must be one of {valid_granularities}.")
                # fmt: on
            return v
        return v

    @U.field_validator("sampling_interval_seconds", mode="before")
    def validate_sampling_interval_seconds(cls, v):
        if v is None:
            return 600
        return v


def build_query_str(query: BaseDataSignalSchema) -> str:
    """
    Build a query string based on the provided BaseDataSignalSchema instance.

    Parameters
    ----------
    query : BaseDataSignalSchema
        The query object containing the filtering criteria.

    Returns
    -------
    str
        The constructed query string.
    """
    if query.sites is None and query.locations is None:
        return "metric.str.contains(@query.metrics, case=False, regex=True)"
    elif query.sites is None and query.locations is not None:
        return (
            "(location.str.lower() == @query.locations or location_id.str.lower() == @query.locations) "
            "and metric.str.contains(@query.metrics, case=False, regex=True)"
        )
    elif query.locations is None and query.sites is not None:
        return (
            "(site.str.lower() == @query.sites or site_id.str.lower() == @query.sites) "
            "and metric.str.contains(@query.metrics, case=False, regex=True)"
        )
    elif (
        query.sites is not None
        and query.locations is not None
        and (query.metrics == ["all"] or query.metrics == "all")
    ):
        return (
            "(site_id.str.lower() == @query.sites or site.str.lower() == @query.sites) "
            "and (location.str.lower() == @query.locations or location_id.str.lower() == @query.locations)"
        )
    else:
        return (
            "(site_id.str.lower() == @query.sites or site.str.lower() == @query.sites) "
            "and (location.str.lower() == @query.locations or location_id.str.lower() == @query.locations) "
            "and metric.str.contains(@query.metrics, case=False, regex=True)"
        )


def get_selected_metrics(
    query: BaseDataSignalSchema,
    metrics_overview: pd.DataFrame,
    log: bool = True,
) -> pd.DataFrame:
    """
    Calculate the selected metrics DataFrame based on the metrics_overview,
    and the query object.

    Parameters
    ----------
    metrics_overview : pd.DataFrame
        The metrics overview DataFrame.
    query_str : str
        The query string to filter metrics.
    log : bool, optional
        Whether to log the selected metrics, by default True.

    Returns
    -------
    pd.DataFrame
        The filtered and sorted metrics DataFrame.
    """
    selected = metrics_overview.query(query.query_str).sort_values(
        ["site", "location", "data_group", "short_hand", "statistic"],
        ascending=[True, True, False, True, True],
    )
    if log:
        logging.info("\033[32;1mMetrics selected for the query:\033[0m\n")
        logging.info(selected[["metric", "unit_str", "site", "location"]])
    if selected.empty:
        raise E.DataSignalsError(
            "\033[31;1mNo metrics found for the given query.\033[0m"
            "\033[33;1m\nHINT:\033[22m Check \033[2msites\033[22m, "
            "\033[2mlocations\033[22m, and \033[2mmetrics\033[22m "
            "provided.\033[0m\n\n"
            "Provided:\n"
            f"  • sites: {query.sites}\n"
            f"  • locations: {query.locations}\n"
            f"  • metrics: {query.metrics}\n"
        )
    return selected


def get_selected_locations(
    query: GetOldestTimestampSchema,
    metrics_overview: pd.DataFrame,
    log: bool = True,
) -> pd.DataFrame:
    """
    Calculate the selected locations DataFrame based on the metrics_overview,
    and the query object.

    Parameters
    ----------
    query: GetOldestTimestampSchema
        Timestamp schema query class instance.
    metrics_overview : pd.DataFrame
        The metrics overview DataFrame.
    log : bool, optional
        Whether to log the selected locations, by default True.

    Returns
    -------
    pd.DataFrame
        The filtered and sorted locations DataFrame.
    """
    # fmt: off
    selected = metrics_overview \
        .query(query.query_str)[["site", "location",
                                 "site_id", "location_id"]] \
        .assign(site_lower=lambda x: x["site"].str.lower(),
                location_lower=lambda x: x["location"].str.lower()) \
        .drop_duplicates(subset=["site_lower", "location_lower"]) \
        .drop(columns=["site_lower", "location_lower"]) \
        .sort_values(["site", "location"], ascending=[True, True])
    # fmt: on
    if log:
        logging.info("\033[32;1mLocations selected for the query:\033[0m\n")
        logging.info(selected[["site", "location"]])
    if selected.empty:
        raise E.DataSignalsError(
            "\033[31;1mNo locations found for the given query.\033[0m"
            "\033[33;1m\nHINT:\033[22m Check \033[2msites\033[22m and "
            "\033[2mlocations\033[22m provided.\033[0m\n\n"
            "Provided:\n"
            f"  • site: {query.site}\n"
            f"  • locations: {query.locations}\n"
        )
    # Check if site_id matches multiple sites
    unique_sites = selected[["site", "site_id"]].drop_duplicates()
    site_id_matches = unique_sites[
        unique_sites["site_id"].str.lower() == query.site
    ]
    if len(site_id_matches) > 1:
        matching_sites = site_id_matches["site"].tolist()
        raise E.DataSignalsError(
            f"\033[31;1msite_id '{query.site}' matches multiple sites: "
            f"{matching_sites}\033[0m\n\033[33;1mHINT:\033[22m Use the site "
            "name directly as input parameter rather than the site_id to avoid "
            "ambiguity.\033[0m\n"
        )
    return selected



# -*- coding: utf-8 -*-
"""Custom exceptions for the 24SEA API."""
import logging
from typing import Optional

from httpx import Request


class AuthenticationError(Exception):
    """An exception to raise when the user is not authenticated."""

    pass


class ProfileError(Exception):
    """An exception to raise when the user is authenticated, but its profile
    is not properly configured."""

    pass


class DataSignalsError(Exception):
    """An exception to raise when the data signals are not properly
    configured."""

    pass


class HTTPError(Exception):
    """
    Base class for ``RequestError`` and ``HTTPStatusError``.

    Useful for  ``try...except`` blocks when issuing a request,
    and then calling ``.raise_for_status()``.

    For example:

    >>> try:
    ...     response = httpx.get("https://www.example.com")
    ...     response.raise_for_status()
    ... except httpx.HTTPError as exc:
    ...     print(f"HTTP Exception for {exc.request.url} - {exc}")
    """

    def __init__(self, message: str, text: str = "") -> None:
        """Initialize the HTTPError."""
        super().__init__(message, text)
        self.text = text
        self._request: Optional[Request] = None

    @property
    def request(self) -> Request:
        if self._request is None:
            raise RuntimeError("The .request property has not been set.")
        return self._request

    @request.setter
    def request(self, request: Request) -> None:
        self._request = request

    def __repr__(self) -> str:
        """Return the representation of the HTTPError."""
        return f"HTTPError({self.args[0]!r}, text={self.text!r})"

    def __str__(self) -> str:
        """Return the string representation of the HTTPError."""
        return str(self.args[0]) + (f" (text={self.text})" if self.text else "")


def raise_for_status(response) -> None:
    """
    Raise an HTTPError if the response status code indicates an error.

    This function checks the status code of the provided HTTP response. If the
    status code is in the 4xx or 5xx range, it raises an HTTPError with a
    descriptive message including the status code and URL.

    Parameters
    ----------
    response : httpx.Response
        The HTTP response object to check for errors.

    Raises
    ------
    HTTPError
        If the response status code is between 400 and 599 (inclusive),
        indicating a client or server error.
    """
    if 400 <= response.status_code < 600:
        http_error_msg = f"\033[1;31m{response.status_code}\n"
        http_error_msg += f"Error for url: {response.url}\033[0m"
        if response.status_code == 500:
            http_error_msg += (
                "\nThe server encountered an internal error and was "
                "unable to complete your request. Please try again later, "
                "or contact support at \033[32;1;4msupport.api@24sea.eu\033[0m "
                "if the problem persists."
            )
        logging.error(http_error_msg)
        raise HTTPError(http_error_msg, text=response.text) from None



# -*- coding: utf-8 -*-
"""Convert cycle-count columns in the DataFrame, if any, to
:class:`py_fatigue.CycleCount` objects. This functionality is available
only if the extra :mod:`py_fatigue` package is installed."""
# from typing import TYPE_CHECKING, NewType

import pandas as pd

from ..singleton import Singleton as S

try:
    swifter = S.swifter
except ImportError:
    pass


try:
    # delete the accessor to avoid warning
    del pd.DataFrame.fatigue  # type: ignore[assignment]
except AttributeError:
    pass


@pd.api.extensions.register_dataframe_accessor("fatigue")
class FatigueAccessor:
    """Accessor for the :mod:`py_fatigue` package."""

    def __init__(self, _obj: pd.DataFrame):
        self._obj = _obj
        self._validate()

    def _validate(self) -> None:
        if S.py_fatigue.CycleCount is None:
            raise ImportError(
                "The `py_fatigue` package is required for this functionality.\n"
                "Reinstall api-24sea with the `fatigue` extra, i.e.,\n"
                "pip install api-24sea[fatigue]"
            )
        # Check that the dataframe contains columns starting with "CC_",
        # otherwise raise a FatigueError
        if not any(col.startswith("CC_") for col in self._obj.columns):
            raise ValueError(
                "The DataFrame does not contain any columns starting with "
                "'CC_', which are required for this functionality."
            )
        # Check that the datafame index is of type DatetimeIndex
        if not isinstance(self._obj.index, pd.DatetimeIndex):
            raise ValueError(
                "The DataFrame index must be of type DatetimeIndex for this "
                "functionality. Currently, the index is of type "
                f"{type(self._obj.index)}."
            )

    def cycle_counts_to_objects(self) -> pd.DataFrame:
        """Convert cycle-count columns in the DataFrame to
        :class:`py_fatigue.CycleCount`"""
        self._validate()
        for col in self._obj.columns:
            if col.startswith("CC_"):
                # fmt: off
                self._obj[col] = self._obj.swifter.progress_bar(
                    desc=(f"\033[32;1mConverting \033[22;34m{col} "
                          "\033[32;1m...")
                ).apply(lambda row: S.py_fatigue.CycleCount.from_rainflow(
                    row[col],
                    name=col[3:],
                    timestamp=row.name)
                 if isinstance(row[col], dict) else row[col], axis=1)
                # fmt: on
        return self._obj



# -*- coding: utf-8 -*-
"""Utility functions and classes."""
import asyncio
import datetime
import logging
import multiprocessing
import sys
import time
import warnings
from collections import defaultdict
from types import CoroutineType
from typing import Any, DefaultDict, Dict, Iterable, List, Optional, Union

import httpx
import pandas as pd
from pydantic import version as pydantic_version
from shorthand_datetime import parse_shorthand_datetime
from tqdm import tqdm
from tqdm.asyncio import tqdm_asyncio
from tqdm.contrib.logging import logging_redirect_tqdm
from tqdm.notebook import tqdm as tqdm_nb

# Local imports
from . import exceptions as E
from . import version


def is_executable_ipython() -> bool:
    """Return True when running under IPython-based shells.

    Returns
    -------
    bool
        True for IPython or Jupyter shells, False for standard Python.

    Examples
    --------
    >>> isinstance(is_executable_ipython(), bool)
    True
    """
    try:
        from IPython import get_ipython  # type: ignore[import-not-found]
    except ImportError:
        return False

    shell = get_ipython()
    if shell is None:
        return False

    shell_name = getattr(shell, "__class__", type(shell)).__name__
    ipy_shells = {
        "ZMQInteractiveShell",
        "TerminalInteractiveShell",
        "PyDevTerminalInteractiveShell",
    }
    if shell_name in ipy_shells:
        return True

    return "ipykernel" in sys.modules


def run_tqdm(
    iterable: Iterable,
    *args: Any,
    exec_env: bool,
    **kwargs: Any,
):
    """Select the appropriate tqdm variant for the active environment.

    Parameters
    ----------
    iterable : Iterable
        Items to iterate over.
    args : Any
        Positional arguments forwarded to ``tqdm``.
    exec_env : bool
        True when the notebook-aware progress bar should be used.
    kwargs : Any
        Keyword arguments forwarded to ``tqdm``.

    Returns
    -------
    tqdm.std.tqdm
        Configured progress iterator.

    Examples
    --------
    >>> list(run_tqdm(range(2), exec_env=False, disable=True))
    [0, 1]
    """
    if exec_env:
        return tqdm_nb(iterable, *args, **kwargs)
    return tqdm(iterable, *args, **kwargs)


BASE_URL = "https://api.24sea.eu/routes/v1/"
PYDANTIC_V2 = version.parse_version(pydantic_version.VERSION).major >= 2

if PYDANTIC_V2:
    from pydantic import field_validator  # type: ignore
    from pydantic import BaseModel, validate_call

else:
    from pydantic import BaseModel, validator  # noqa: F401

    # Fallback for validate_call (acts as a no-op)
    def validate_call(*args, **kwargs):
        # Remove config kwarg if present since it's not supported in v1
        if "config" in kwargs:
            del kwargs["config"]

        def decorator(func):
            return func

        if args and callable(args[0]):
            return decorator(args[0])
        return decorator

    # Shim for field_validator to behave like validator
    def field_validator(field_name, *args, **kwargs):
        def decorator(func):
            # Convert mode='before' to pre=True for v1 compatibility
            if "mode" in kwargs:
                if kwargs["mode"] == "before":
                    kwargs["pre"] = True
                del kwargs["mode"]
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                return validator(field_name, *args, **kwargs)(func)

        return decorator


def handle_request(
    url: str,
    params: Dict,
    auth: Optional[httpx.BasicAuth],
    headers: Optional[Dict[str, str]] = {"accept": "application/json"},
    max_retries: int = 10,
    timeout: int = 3600,
) -> httpx.Response:
    """Handle the request to the 24SEA API and manage errors using httpx.

    This function will handle the request to the 24SEA API and manage any
    errors that may arise. If the request is successful, the response object
    will be returned. Otherwise, an error will be raised.

    Parameters
    ----------
    url : str
        The URL to which to send the request.
    params : dict
        The parameters to send with the request.
    auth : httpx.BasicAuth
        The authentication object.
    headers : dict
        The headers to send with the request.

    Returns
    -------
    httpx.Response
        The response object if the request was successful, otherwise error.
    """
    if auth is None:
        auth = httpx.BasicAuth("", "")
    retry_count = 0

    while True:
        try:
            # fmt: off
            r_ = httpx.get(url, params=params, auth=auth, headers=headers,
                           timeout=timeout)
            # fmt: on
            if r_.status_code != 502 or retry_count >= max_retries:
                break
            retry_count += 1
            if retry_count <= max_retries:
                time.sleep(3)
                continue
        except httpx.RequestError as exc:
            # Bubble up network/timeout errors for caller to handle
            raise exc
    # Validate status after finishing retry loop
    E.raise_for_status(r_)
    return r_


async def handle_request_async(
    url: str,
    params: Dict,
    auth: Optional[httpx.BasicAuth],
    headers: Optional[Dict[str, str]] = {"accept": "application/json"},
    max_retries: int = 10,
    timeout: int = 1800,
    method: str = "GET",
    json: Optional[Dict] = None,
) -> httpx.Response:
    """Asynchronously handle the request to the 24SEA API using httpx's
    AsyncClient. Supports GET, POST, PUT, DELETE methods."""
    retry_count = 0
    async with httpx.AsyncClient(
        auth=auth, headers=headers, timeout=timeout
    ) as client:
        while True:
            try:
                if method.upper() == "GET":
                    r_ = await client.get(url, params=params)
                elif method.upper() == "POST":
                    r_ = await client.post(url, params=params, json=json)
                elif method.upper() == "PUT":
                    r_ = await client.put(url, params=params, json=json)
                elif method.upper() == "DELETE":
                    r_ = await client.delete(url, params=params)
                else:
                    raise ValueError(f"Unsupported HTTP method: {method}")
                if r_.status_code != 502 or retry_count >= max_retries:
                    break
                retry_count += 1
                if retry_count <= max_retries:
                    await asyncio.sleep(3)
                    continue
            except (httpx.NetworkError, httpx.TimeoutException) as exc:
                raise exc
        E.raise_for_status(r_)
        return r_


def default_to_regular_dict(d_: Union[DefaultDict, Dict]) -> Dict:
    """Convert a defaultdict to a regular dictionary."""
    if isinstance(d_, defaultdict):
        return {k: default_to_regular_dict(v) for k, v in d_.items()}
    return d_


def require_auth(func):
    """Decorator to ensure authentication before executing a method"""

    def wrapper(self, *args, **kwargs):
        """Wrapper function to check authentication."""
        if not self.authenticated:
            self._lazy_authenticate()
        if not self.authenticated:
            raise E.AuthenticationError(
                "\033[31;1mAuthentication needed before querying the metrics.\n"
                "\033[0mUse the \033[34;1mdatasignals.\033[0mauthenticate("
                "\033[32;1m<username>\033[0m, \033[32;1m<password>\033[0m) "
                "method."
            )
        return func(self, *args, **kwargs)

    return wrapper


# def require_auth_async(func):
#     """Decorator to ensure authentication before executing a method"""

#     async def wrapper(self, *args, **kwargs):
#         """Wrapper function to check authentication."""
#         if not self.authenticated:
#             await self._lazy_authenticate()
#         if not self.authenticated:
#             raise E.AuthenticationError(
#                 "\033[31;1mAuthentication needed before querying the metrics.\n"
#                 "\033[0mUse the \033[34;1mdatasignals.\033[0mauthenticate("
#                 "\033[32;1m<username>\033[0m, \033[32;1m<password>\033[0m) "
#                 "method."
#             )
#         return await func(self, *args, **kwargs)

#     return wrapper


def parse_timestamp(
    df: pd.DataFrame,
    formats: Iterable[str] = ("ISO8601", "mixed"),
    dayfirst: bool = False,
    keep_index_only: bool = True,
) -> pd.DataFrame:
    """Parse timestamp column in DataFrame using multiple format attempts.

    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame containing timestamp column or index
    formats : Iterable[str], default ('ISO8601', 'mixed')
        List of datetime format strings to try
    dayfirst : bool, default False
        Whether to interpret dates as day first

    Returns
    -------
    pandas.DataFrame
        DataFrame with parsed timestamp column

    Raises
    ------
    ValueError
        If timestamp parsing fails with all formats
    """
    series = None
    d_e = (
        f"No format matched the timestamp index/column among {formats}.\n"
        "            Try calling `parse_timestamp` manually with another "
        "format, e.g.,\n"
        "            \033[32;1m>>>\033[31;1m import\033[0m api_24sea.utils "
        "\033[31;1mas \033[0mU\n"
        "            \033[32;1m>>>\033[0m U.parse_timestamp(df,\n"
        "                                  formats=\033[32m[\033[36m"
        "'YYYY-MM-DDTHH:MM:SSZ'\033[32m]\033[0m,\n"
        "                                  dayfirst=\033[34mFalse\033[0m)"
    )

    if df.index.name == "timestamp":
        if "timestamp" in df.columns:
            # fmt: off
            logging.warning("Both index and column named 'timestamp' found. "
                            "Index takes precedence.")
            # fmt: on
            # Drop the column if it's not the index
            df.drop(columns="timestamp", inplace=True)
        series = df.index.to_series()
    else:
        if "timestamp" in df.columns:
            if df["timestamp"].isnull().all():
                # fmt: off
                raise E.DataSignalsError("`data` must include a 'timestamp' "
                                         "column or indices convertible to "
                                         "timestamps.")
                # fmt: on
            series = df["timestamp"]
    if series is None:
        raise E.DataSignalsError(d_e)
    try:
        # Try parsing with different formats
        for fmt in formats:
            try:
                df["timestamp"] = pd.to_datetime(
                    series, format=fmt, dayfirst=dayfirst, errors="raise"
                )
                if keep_index_only:
                    df.set_index("timestamp", inplace=True)
                return df
            except ValueError:
                continue
        # fmt: off
        # If all previous attempts failed, it means that pandas version
        # is not compatible with the formats provided, therefore try
        # with the following formats.
        formats = ["%Y-%m-%dT%H:%M:%S%z", "%d.%m.%YT%H:%M:%S.%f%z",
                   "%Y-%m-%dT%H:%M:%SZ", "%d.%m.%YT%H:%M:%S.%fZ", "%Y-%m"]
        # fmt: on
        df["timestamp"] = pd.NaT
        for fmt in formats:
            temp_series = pd.to_datetime(series, format=fmt, errors="coerce")
            df["timestamp"].fillna(temp_series, inplace=True)
        if keep_index_only:
            df.set_index("timestamp", inplace=True)
        return df
    except Exception as exc:
        logging.error(f"All timestamp parsing attempts failed: {str(exc)}")
        raise E.DataSignalsError("Could not parse timestamp data") from exc


def estimate_chunk_size(
    tasks: list,
    start_timestamp: Union[str, datetime.datetime],
    end_timestamp: Union[str, datetime.datetime],
    grouped_metrics: Iterable,
    selected_metrics: Union[pd.DataFrame, None] = None,
    target: str = "metric",
):
    """
    Estimate the optimal chunk size for processing tasks based on the expected
    data volume.
    This function calculates the estimated size of the data request in megabytes
    (MB) by considering the number of data points, the number of tasks, and the
    bytes required per metric. It then determines an appropriate chunk size for
    processing the tasks efficiently.

    Parameters
    ----------
    tasks : list
        List of tasks to be processed.
    query : object
        Query object containing at least `start_timestamp` and `end_timestamp`
        attributes.
    grouped_metrics : iterable
        Iterable of grouped metrics, where each group is a tuple (key, group),
        and group is typically a DataFrame.
    selected_metrics : pandas.DataFrame or None
        DataFrame containing selected metrics with at least a "metric" column
        and optionally a "data_group" column.
    target : str, default "metric"
        The target column name in `selected_metrics` and `grouped_metrics`

    Returns
    -------
    dict
        Dictionary with the following keys:
            - "total_mb": float, estimated total size of the request in MB.
            - "n_tasks": int, number of tasks.
            - "chunk_size": int, recommended chunk size for processing.

    Notes
    -----
    - The function assumes each data point is a float64 (8 bytes) unless
      overridden by the "data_group".
    - The number of data points is estimated as one every 10 minutes between the
      start and end timestamps.
    - Chunk size is determined based on the estimated total data size.
    """
    logging.debug("Estimating chunk size...")
    logging.debug(f"Number of tasks: {len(tasks)}")
    logging.debug(f"Start timestamp: {start_timestamp}")
    logging.debug(f"End timestamp: {end_timestamp}")
    logging.debug(f"Number of grouped metrics: {len(list(grouped_metrics))}")
    logging.debug(f"Grouped metrics: {list(grouped_metrics)}")
    logging.debug(f"Selected metrics:\n{selected_metrics}")

    def parse_dt(dt):
        if isinstance(dt, str):
            try:
                return pd.to_datetime(dt)
            except pd._libs.tslibs.parsing.DateParseError:  # type: ignore
                _dt = parse_shorthand_datetime(dt)
                if _dt is not None:
                    dt = _dt.replace(tzinfo=None)
        return dt

    start_dt = parse_dt(start_timestamp)
    end_dt = parse_dt(end_timestamp)
    n_minutes = (end_dt - start_dt).total_seconds() / 60  # type: ignore
    n_points = int(n_minutes // 10) + 1
    n_tasks = len(tasks)
    # Build a dictionary of bytes per metric
    bytes_per_metric = {}
    if selected_metrics is not None:
        for _, row in selected_metrics.iterrows():
            metric = row[target]
            data_group = str(row.get("data_group", "")).lower()
            bytes_per_metric[metric] = 8
            if data_group == "fatigue":
                bytes_per_metric[metric] *= 25
            if data_group == "mpe":
                bytes_per_metric[metric] *= 7
            if data_group == "mdl":
                bytes_per_metric[metric] *= 7
            if "_flt_" in metric.lower():
                bytes_per_metric[metric] *= 35
    total_bytes = 0
    for _, group in grouped_metrics:
        if isinstance(group, pd.DataFrame):
            group_met = group[target].tolist()
        else:
            group_met = [group[target]] if hasattr(group, target) else []
        # If group_met contains only "all", use all selected metrics
        if group_met == ["all"] and selected_metrics is not None:
            group_met = selected_metrics[target].tolist()
        group_bytes = sum(bytes_per_metric.get(m, 8) for m in group_met)
        total_bytes += n_points * group_bytes

    # Check for negative or zero values (sanity check)
    if total_bytes <= 0 or n_points <= 0:
        total_bytes = 0
    total_mb = total_bytes / (1024 * 1024)

    # Determine chunk_size
    if total_mb < 40:
        chunk_size = n_tasks
    elif total_mb < 80:
        chunk_size = max(1, n_tasks // 2)
    elif total_mb < 160:
        chunk_size = max(1, n_tasks // 4)
    else:
        chunk_size = max(1, n_tasks // 8)

    logging.info(f"Estimated request size: {total_mb:.2f} MB")
    return {
        "total_mb": total_mb,
        "n_tasks": n_tasks,
        "chunk_size": chunk_size,
    }


async def gather_in_chunks(
    tasks: List[CoroutineType], chunk_size: int = 5, timeout: int = 3600
) -> List:
    results = []
    chunk_results = []
    with logging_redirect_tqdm():
        total_tasks = len(tasks)

        if total_tasks == 1:
            desc = "API-24SEA get_data"
        elif chunk_size == 1:
            desc = f"API-24SEA get_data [total locations: {total_tasks}]"
        else:
            # fmt: off
            desc = (f"API-24SEA get_data in {chunk_size}-sized chunks "
                    f"[total locations: {total_tasks}]")
        exec_env = is_executable_ipython()
        for i in run_tqdm(
            range(0, len(tasks), max(1, chunk_size)),
            exec_env=exec_env,
            desc=desc,
            colour="#c9cfd8",
            position=0,
        ):
            chunk = tasks[i : i + chunk_size]
            chunk_results = await tqdm_asyncio.gather(
                *chunk,
                desc=f"Getting chunk: [{i+1}-{i+len(chunk)}]",
                timeout=timeout,
                colour="#e4e8ee",
                position=1,
            )
            results.extend(chunk_results)
    return results


def fetch_data_sync(
    url,
    site: str,
    location: str,
    start_timestamp: Union[datetime.datetime, str],
    end_timestamp: Union[datetime.datetime, str],
    headers: Optional[Dict[str, str]],
    group: pd.DataFrame,
    auth: Optional[httpx.BasicAuth],
    timeout: int,
    target: str = "metric",
) -> Any:
    """Syncronously fetch metrics data for the datasignals API app."""
    # fmt: off
    s_ = "• " + ",".join(group[target].tolist()).replace(
        ",", f"\n {len(target) + 1}    • "
    )
    logging.info(f"\033[32;1m⏳ Getting data for {site} - "
                    f"{location}...\n📊 \033[35;1m{target.capitalize()}s: "
                    f"\033[0;34m{s_}\n\033[0m")
    # fmt: on
    r_ = handle_request(
        url,
        {
            "start_timestamp": start_timestamp,
            "end_timestamp": end_timestamp,
            "project": [site],
            "location": [location],
            f"{target}s": ",".join(group[target].tolist()),
        },
        auth,
        headers,
        timeout=timeout,
    )
    # Warn if empty
    if r_.json() == []:
        logging.warning(
            f"\033[33;1m⚠️ No data found for {site} - " f"{location}.\033[0m"
        )
    return r_.json()


def fetch_availability_sync(
    url,
    site: str,
    location: str,
    start_timestamp: Union[datetime.datetime, str],
    end_timestamp: Union[datetime.datetime, str],
    granularity: Union[str, int],
    sampling_interval_seconds: int,
    headers: Optional[Dict[str, str]],
    group: pd.DataFrame,
    auth: Optional[httpx.BasicAuth],
    timeout: int,
) -> Any:
    """Syncronously fetch metrics data for the datasignals API app."""
    # fmt: off
    s_ = "• " + ",".join(group["metric"].tolist()).replace(",", "\n            • ")  # noqa: E501  # pylint: disable=C0301
    logging.info(f"\033[32;1m⏳ Getting availabilities for {site} - "
                    f"{location}...\n📊 \033[35;1mMetrics: "
                    f"\033[0;34m{s_}\n\033[0m")
    params: Dict[str, Union[List[str], datetime.datetime, str, int]] = {
        "start_timestamp": start_timestamp,
        "end_timestamp": end_timestamp,
        "project": [site],
        "location": [location],
        "metrics": ",".join(group["metric"].tolist()),
    }
    # fmt: on
    if isinstance(granularity, str):
        params["granularity"] = granularity
    if isinstance(granularity, int):
        params["bucket_seconds"] = granularity
    params["sampling_interval_seconds"] = sampling_interval_seconds

    r_ = handle_request(url, params, auth, headers, timeout=timeout)
    # Warn if empty
    if r_.json() == []:
        logging.warning(
            f"\033[33;1m⚠️ No data found for {site} - " f"{location}.\033[0m"
        )
    return r_.json()


async def fetch_availability_async(
    url: str,
    site: str,
    location: str,
    start_timestamp: Union[datetime.datetime, str],
    end_timestamp: Union[datetime.datetime, str],
    granularity: Union[str, int],
    sampling_interval_seconds: int,
    headers: Optional[Dict[str, str]],
    group: pd.DataFrame,
    auth: Optional[httpx.BasicAuth],
    timeout: int,
    max_retries: int,
) -> Any:
    """Asynchronously fetch availability data for a site/location."""
    s_ = "• " + ",".join(group["metric"].tolist()).replace(
        ",", "\n            • "
    )
    logging.info(
        f"\033[32;1m⏳ Getting availabilities for {site} - {location}..."
        f"\n📊 \033[35;1mMetrics: \033[0;34m{s_}\n\033[0m"
    )
    params: Dict[str, Union[List[str], datetime.datetime, str, int]] = {
        "start_timestamp": start_timestamp,
        "end_timestamp": end_timestamp,
        "project": [site],
        "location": [location],
        "metrics": ",".join(group["metric"].tolist()),
    }
    if isinstance(granularity, str):
        params["granularity"] = granularity
    else:
        params["bucket_seconds"] = granularity
    params["sampling_interval_seconds"] = sampling_interval_seconds
    r_ = await handle_request_async(
        url, params, auth, headers, max_retries=max_retries, timeout=timeout
    )
    result_json = r_.json()
    if result_json == []:
        logging.warning(
            f"\033[33;1m⚠️ No data found for {site} - {location}.\033[0m"
        )
    return result_json


async def fetch_data_async(
    url,
    site: str,
    location: str,
    start_timestamp: Union[datetime.datetime, str],
    end_timestamp: Union[datetime.datetime, str],
    headers: Optional[Dict[str, str]],
    group: pd.DataFrame,
    auth: Optional[httpx.BasicAuth],
    timeout: int,
    max_retries: int,
    as_dict: bool = False,
    target: str = "metric",
) -> Union[pd.DataFrame, Dict[str, Any]]:
    """Asyncronously fetch metrics data for the datasignals API app."""
    tgt_str = "data" if target == "metric" else "predictions"
    s_ = "• " + ",".join(group[target].tolist()).replace(
        ",", "\n            • "
    )
    logging.info(
        f"\033[32;1m⏳ Getting {tgt_str} for {site} - {location}..."
        f"\n📊 \033[35;1m{target.capitalize()}s: \033[0;34m{s_}\n\033[0m"
    )
    r_ = await handle_request_async(
        url,
        {
            "start_timestamp": start_timestamp,
            "end_timestamp": end_timestamp,
            "project": [site],
            "location": [location],
            f"{target}s": ",".join(group[target].tolist()),
        },
        auth,
        headers,
        max_retries=max_retries,
        timeout=timeout,
    )
    result_json = r_.json()
    if result_json == []:
        logging.warning(
            f"\033[33;1m⚠️ No {tgt_str} found for {site} - {location}.\033[0m"
        )
    if as_dict:
        return result_json
    return pd.DataFrame(result_json)


async def fetch_oldest_timestamp_async(
    url,
    site: str,
    locations: Optional[str],
    headers: Dict[str, str],
    auth: Optional[httpx.BasicAuth],
    timeout: int,
    max_retries: int,
    as_dict: bool = False,
) -> Union[pd.DataFrame, Dict[str, Any]]:
    # Format locations with bullets and newlines
    if locations:
        formatted_locations = "   • " + "\n   • ".join(locations.split(","))
        logging.info(
            f"\033[32;1m⏳ Getting oldest timestamps for {site} at the "
            f"following locations:\n{formatted_locations}\n\033[0m"
        )
    else:
        logging.info(
            f"\033[32;1m⏳ Getting oldest timestamps for {site}\n\033[0m"
        )
    r_ = await handle_request_async(
        url,
        (
            {"project": site, "locations": locations}
            if locations
            else {"project": site}
        ),
        auth,
        headers,
        max_retries=max_retries,
        timeout=timeout,
    )
    result_json = r_.json()
    if result_json == []:
        logging.warning(f"\033[33;1m⚠️ No data found for {site}.\033[0m")
    if as_dict:
        return result_json
    return pd.DataFrame(result_json)


def set_threads_nr(threads: Optional[int], thread_limit: int = 30) -> int:
    """
    Set the number of threads to use for processing.

    Parameters
    ----------
    threads : Optional[int]
        The number of threads to use. If None, the number of available CPU cores
        will be used.

    Returns
    -------
    int
        The number of threads to use.
    """
    if threads is None:
        return multiprocessing.cpu_count()
    if threads < 1:
        return 1
    return int(threads) if threads < thread_limit else thread_limit


def parse_stats_list(stats_list: List[Dict[str, Any]]) -> pd.DataFrame:
    """
    Parse a list of statistics dictionaries into a DataFrame.

    Parameters
    ----------
    stats_list : List[Dict[str, Any]]
        List of dictionaries containing statistics data.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the parsed statistics.
    """
    if not stats_list:
        return pd.DataFrame()

    rows = []
    for d in stats_list:
        for k, v in d.items():
            if "_" in k:
                prefix, metric = k.split("_", 1)
                rows.append({"metric": metric, "stat_type": prefix, "value": v})

    if not rows:
        return pd.DataFrame()

    # Create DataFrame and pivot in one operation
    df = pd.DataFrame(rows)
    stats_df = df.pivot_table(
        index="metric", columns="stat_type", values="value", aggfunc="first"
    ).reset_index()

    # Reorder and clean
    raw_prefixes = sorted(c_ for c_ in stats_df.columns if c_ != "metric")
    cols_order = ["metric"] + sorted(raw_prefixes)
    for c in cols_order:
        if c not in stats_df.columns:
            stats_df[c] = None
    stats_df = stats_df[cols_order]

    if isinstance(stats_df, pd.Series):
        stats_df = stats_df.to_frame().T  # Convert Series to DataFrame

    return stats_df.reset_index(drop=True)


def get_stats_overview_info(
    stats_df: pd.DataFrame,
    metrics_overview: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Get the overview information for statistics DataFrame.

    Parameters
    ----------
    stats_df : pd.DataFrame
        DataFrame containing statistics data.
    metrics_overview : pd.DataFrame
        DataFrame containing metrics overview information.

    Returns
    -------
    pd.DataFrame
        DataFrame with overview information merged with stats_df.
    """
    if metrics_overview is None:
        return stats_df
    if stats_df.empty or metrics_overview.empty:
        return stats_df

    stats_df_temp = stats_df.copy()
    overview_temp = metrics_overview.copy()
    stats_df_temp["metric_lower"] = stats_df_temp["metric"].str.lower()
    overview_temp["metric_lower"] = overview_temp["metric"].str.lower()
    merged = stats_df_temp.merge(
        overview_temp[
            [
                "metric",
                "metric_lower",
                "site",
                "location",
                "data_group",
                "statistic",
                "short_hand",
                "print_str",
            ]
        ],
        on="metric_lower",
        how="left",
        suffixes=("", "_overview"),
    )
    merged["metric"] = merged["metric_overview"].fillna(merged["metric"])
    return merged.drop(["metric_lower", "metric_overview"], axis=1)


def get_stats_as_dict(
    stats_df: pd.DataFrame,
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """
    Convert the statistics DataFrame to a dictionary format.

    Parameters
    ----------
    stats_df : pd.DataFrame
        DataFrame containing statistics data.

    Returns
    -------
    Dict[str, Dict[str, pd.DataFrame]]
        Dictionary with site and location as keys and statistics as values.
    """
    result_dict: DefaultDict[str, DefaultDict[str, pd.DataFrame]] = defaultdict(
        lambda: defaultdict(pd.DataFrame)
    )
    # fmt: off
    for (s_, l_), group in stats_df.groupby(['site', 'location']):  # type: ignore  # pylint: disable=C0301  # noqa: E501
        s_ = s_.lower()  # type: ignore
        l_ = l_.upper()  # type: ignore
        result_dict[s_][l_] = group.drop(['site', 'location'],
                                            axis=1).reset_index(drop=True)
    # fmt: on
    return default_to_regular_dict(result_dict)


def get_metrics_data_df_as_dict(
    metrics_data_df: pd.DataFrame, selected_metrics: pd.DataFrame
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """
    Convert the metrics DataFrame to a dictionary format.

    Parameters
    ----------
    metrics_data_df : pd.DataFrame
        DataFrame containing metrics data.

    Returns
    -------
    Dict[str, Dict[str, pd.DataFrame]]
        Dictionary with site and location as keys and metrics data as values.
    """
    # We need to reset the dataframe index to get the timestamp column
    # as a column and not as the index.
    metrics_data_df = parse_timestamp(metrics_data_df, keep_index_only=False)
    # fmt: off
    __melted = pd.melt(metrics_data_df, id_vars=["timestamp"],
                        var_name="metric", value_name="value")
    # Case-insensitive merge on metric (keep original 'metric' columns, only drop 'metric_lower')
    __melted["metric_lower"] = __melted["metric"].str.lower()
    _selected_metrics_ci = selected_metrics.copy()
    _selected_metrics_ci["metric_lower"] = _selected_metrics_ci["metric"].str.lower()
    __merged = pd.merge(__melted, _selected_metrics_ci, on="metric_lower", how="left")

    # Resolve possible duplicate metric columns (metric_x from melted, metric_y from selected)
    if "metric_x" in __merged.columns and "metric_y" in __merged.columns:
        # Prefer the casing from the selected metrics dataframe when available
        __merged["metric"] = __merged["metric_y"].fillna(__merged["metric_x"])
        __merged.drop(columns=[c for c in ["metric_x", "metric_y",
                                           "metric_lower"]
                               if c in __merged.columns], inplace=True)
    else:
        # Just drop helper column
        if "metric_lower" in __merged.columns:
            __merged.drop(columns=["metric_lower"], inplace=True)
    __srt = (__merged[["timestamp", "site", "location", "metric", "value"]]  # type: ignore  # pylint: disable=C0301  # noqa: E501
             .sort_values(by=["timestamp", "location"]).reset_index(drop=True))
    __df = (__srt.pivot_table(index=["site", "location", "timestamp"],
                              columns="metric", values="value",
                              aggfunc="first",).reset_index())
    # fmt: on
    if __df.empty:
        raise E.DataSignalsError(
            "\033[31mThe \033[1mas_dict \033[22mmethod can only be called "
            "when the DataFrame is not empty."
        )
    if not all(c_ in __df.columns for c_ in ["site", "location", "timestamp"]):
        raise E.DataSignalsError(
            "\033[31mThe \033[1mas_dict \033[22mmethod can only be called "
            "when the site, location, and timestamp columns are available."
        )
    if not isinstance(__df.index, pd.RangeIndex):
        raise E.DataSignalsError(
            "\033[31mThe \033[1mas_dict \033[22mmethod can only be called "
            "when the index is a \033[1mRangeIndex.\n"
            "\033[22mThe index is currently a \033[1m"
            f"{type(__df.index).__name__}.\n"
        )
    # Group the DataFrame by site and location
    groups = __df.groupby(["site", "location"])
    __dict: Dict[str, Dict[str, pd.DataFrame]] = {}
    for (s_, l_), group in groups:
        s_ = s_.lower()
        l_ = l_.upper()
        if s_ not in __dict:
            __dict[s_] = {}
        # Manipulate the group to remove columns with all NaN values
        # and set the index to the timestamp column.
        _df: pd.DataFrame = group.drop(["site", "location"], axis=1).dropna(
            axis=1, how="all"
        )
        _df = parse_timestamp(_df)
        _df = _df.sort_index()  # sort by timestamp
        if l_ not in __dict[s_]:
            __dict[s_][l_] = _df

    return default_to_regular_dict(__dict)


def series_to_type(
    series: pd.Series, dtype: Union[str, type]
) -> Union[pd.Series, pd.Timestamp]:
    """
    Convert a pandas Series to a specified data type.

    Parameters
    ----------
    series : pd.Series
        The Series to convert.
    dtype : Union[str, type]
        The data type to convert the series to.

    Returns
    -------
    pd.Series
        The Series converted to the specified data type.

    Example
    -------
    >>> import pandas as pd
    >>> s = pd.Series([1, 2, 3])
    >>> column_to_type(s, float)
    0    1.0
    1    2.0
    2    3.0
    dtype: float64
    """
    if dtype == "datetime":
        return pd.to_datetime(series)
    if dtype in ("float", float, "int", int):
        return pd.to_numeric(series, errors="coerce").astype(dtype)  # type: ignore  # pylint: disable=C301  # noqa: E501
    if dtype in ("str", "string", str):
        return series.astype(str)
    return series


def column_to_type(
    data: pd.DataFrame, column: str, dtype: Union[str, type]
) -> pd.DataFrame:
    """
    Convert a column in a DataFrame to a specified data type.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame containing the column to convert.
    column : str
        The column to convert.
    dtype : Union[str, type]
        The data type to convert the column to.

    Returns
    -------
    pd.DataFrame
        The DataFrame with the column converted to the specified data type.

    Example
    -------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'A': [1, 2, 3], 'B': [4, 5, 6]})
    >>> column_to_type(df, 'A', float)
       A  B
    0  1  4
    1  2  5
    2  3  6
    """
    data[column] = series_to_type(data[column], dtype)
    return data


def calendar_monthly_availability(
    df: pd.DataFrame,
    *,
    start_timestamp: Union[datetime.datetime, str, None] = None,
    end_timestamp: Union[datetime.datetime, str, None] = None,
    sampling_interval_seconds: int = 600,
) -> pd.DataFrame:
    """Convert a daily availability dataframe to a calendar monthly availability
    dataframe.
    The columns of the input dataframe are assumed to be daily availability
    values (between 0 and 1). The output dataframe will have the same columns,
    but the two indices will be year and month. The values will be the mean of
    the daily availability values in that month.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe with daily availability values.

    Returns
    -------
    pd.DataFrame
        Output dataframe with calendar monthly availability values.

    Example
    -------
    >>> import pandas as pd
    >>> data = {'timestamp': pd.date_range(start='2023-01-01', periods=90,
    ...                                    freq='D'),
    ...         'availability': [0.9, 0.8, 0.95] * 30}
    >>> df = pd.DataFrame(data).set_index('timestamp')
    >>> calendar_monthly_availability(df)
                availability
    timestamp
    2023-01          0.883333
    2023-02          0.883333
    2023-03          0.883333
    """
    df = parse_timestamp(df)
    if df.empty:
        return df

    idx = df.index
    tz = idx.tz if isinstance(idx, pd.DatetimeIndex) else None

    range_start = (
        pd.to_datetime(start_timestamp)
        if start_timestamp is not None
        else idx.min()
    )
    range_end = (
        pd.to_datetime(end_timestamp)
        if end_timestamp is not None
        else idx.max()
    )

    if tz is not None:
        range_start = (
            range_start.tz_convert(tz)
            if range_start.tzinfo is not None
            else range_start.tz_localize(tz)
        )
        range_end = (
            range_end.tz_convert(tz)
            if range_end.tzinfo is not None
            else range_end.tz_localize(tz)
        )

    start_day = range_start.floor("D")
    end_day = (
        (range_end - pd.Timedelta(nanoseconds=1)).floor("D")
        if end_timestamp is not None
        else range_end.floor("D")
    )
    full_days = pd.date_range(start=start_day, end=end_day, freq="D", tz=tz)
    full_days = full_days.rename("timestamp")

    # Fill missing days with 0 availability so expected samples are not based
    # on how many daily rows happen to be present.
    df_daily = df.reindex(full_days).fillna(0)

    expected_per_day = pd.Timedelta(days=1) / pd.Timedelta(
        seconds=sampling_interval_seconds
    )
    # Convert expected-per-day to an integer when possible to avoid float noise.
    expected_per_day = (
        int(expected_per_day)
        if float(expected_per_day).is_integer()
        else float(expected_per_day)
    )

    # Convert daily availability ratios into expected/actual sample counts.
    monthly_actual = (df_daily * expected_per_day).resample("MS").sum()
    monthly_days = pd.Series(1, index=full_days).resample("MS").sum()
    monthly_expected = monthly_days * expected_per_day

    df_monthly = monthly_actual.div(monthly_expected, axis=0)

    # Keep backward-compatible index semantics:
    # - strip timezone info (older implementation returned tz-naive)
    # - drop freq metadata to avoid brittle equality checks
    idx_out = df_monthly.index
    if isinstance(idx_out, pd.DatetimeIndex):
        if idx_out.tz is not None:
            idx_out = idx_out.tz_convert(None)
        idx_out = pd.DatetimeIndex(idx_out.to_numpy(), name=idx_out.name)
        df_monthly.index = idx_out

    return df_monthly



# -*- coding: utf-8 -*-
# noqa: E501
# pylint: disable=C0301
"""Facilitate the users' interaction with the 24SEA API (https://api/24sea/eu) by providing pandas interfaces to the API endpoints.

The subpackages are:

- :mod:`api_24sea.datasignals`: Contains:

  * :class:`api_24sea.datasignals.DataSignals` class, which is an accessor for
    transforming data signals from the 24SEA API into pandas DataFrames.
  * :mod:`api_24sea.datasignals.core`: Contains the core :class:`API` and
    :class:`AsyncAPI` classes, providing synchronous and asynchronous
    interfaces for interacting with the 24SEA API - DataSignals app.
    Serialization/deserialization methods are also provided to save and
    load API instances with their authentication state.
  * :mod:`api_24sea.datasignals.fatigue`: Contains the :class:`FatigueAccessor`
    class, which is a pandas DataFrame accessor for converting cycle-count
    columns into :class:`py_fatigue.CycleCount` objects.
  * :mod:`api_24sea.datasignals.schemas`: Contains pydantic models for
    validating input parameters.

The submodules are:

- :mod:`api_24sea.abc`: Contains the AuthABC abstract base class, which provides
  a common interface for authentication classes.
- :mod:`api_24sea.exceptions`: Contains custom exceptions for the package.
- :mod:`api_24sea.singleton`: Contains the :class:`Singleton` class, which
  is a metaclass to facilitate the implementation usage with and without
  extras.
- :mod:`api_24sea.utils`: Contains utility functions and classes to help
  manage requests to the 24SEA API.
- :mod:`api_24sea.version`: Contains the version number of the package.
"""

from . import datasignals

__all__ = ["datasignals"]

from pkgutil import extend_path

from .version import __version__ as __version__  # noqa: F401

__path__ = extend_path(__path__, __name__)  # type: ignore[name-defined]



# -*- coding: utf-8 -*-
"""Abstract base class for authentication mixin."""

import logging
import os
from typing import Optional
from warnings import simplefilter

import httpx
import pandas as pd

from . import utils as U

simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

logging.basicConfig(format="%(message)s", level=logging.WARNING)


class AuthABC:
    """Abstract base class for authentication mixin."""

    def __init__(self):
        self.base_url: str = f"{U.BASE_URL}"
        self._username: Optional[str] = None
        self._password: Optional[str] = None
        self._auth: Optional[httpx.BasicAuth] = None
        self._authenticated: bool = False
        self._permissions_overview: Optional[pd.DataFrame] = None

    def _set_auth(
        self, username: Optional[str], password: Optional[str]
    ) -> None:
        """Set the authentication credentials."""
        self._username = username
        self._password = password
        self._auth = httpx.BasicAuth(self._username, self._password)  # type: ignore  # pylint: disable=C0301  # noqa:E501
        self._authenticated = False

    def _clear_auth(self) -> None:
        """Clear the authentication credentials."""
        self._username = None
        self._password = None
        self._auth = None
        self._authenticated = False

    def _is_authenticated(self) -> bool:
        """Check if the client is authenticated."""
        return self._authenticated

    def _authenticate(self, username: str, password: str) -> bool:
        """Authenticate the client with the given credentials."""
        self._set_auth(username, password)
        self._authenticated = True
        return True

    @property
    def authenticated(self) -> bool:
        """Whether the client is authenticated"""
        return self._authenticated

    def _lazy_authenticate(self) -> bool:
        """Attempt authentication using environment variables"""
        if self._username and self._password:
            self.authenticate(self._username, self._password)
            return True
        username = (
            os.getenv("API_24SEA_USERNAME")
            or os.getenv("24SEA_API_USERNAME")
            or os.getenv("TWOFOURSEA_API_USERNAME")
            or os.getenv("API_TWOFOURSEA_USERNAME")
        )
        password = (
            os.getenv("API_24SEA_PASSWORD")
            or os.getenv("24SEA_API_PASSWORD")
            or os.getenv("TWOFOURSEA_API_PASSWORD")
            or os.getenv("API_TWOFOURSEA_PASSWORD")
        )
        if username and password:
            self.authenticate(username, password)
            return True
        return False

    @property
    @U.require_auth
    def permissions_overview(self) -> Optional[pd.DataFrame]:
        """Get the permissions overview DataFrame."""
        return self._permissions_overview

    @U.validate_call(config=dict(arbitrary_types_allowed=True))
    def authenticate(
        self,
        username: str,
        password: str,
        permissions_overview: Optional[pd.DataFrame] = None,
    ):
        """Meta for authenticate method"""
        self._permissions_overview = (
            permissions_overview
            if permissions_overview is not None
            else self._permissions_overview
        )
        return self
